#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import json
import re
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Any, Generic, Literal, TextIO, TypeGuard, TypeVar
from urllib.parse import urlparse

from rich.console import Console

from otterdog.logging import _print_message, get_logger, is_info_enabled

if TYPE_CHECKING:
    from argparse import Namespace
    from collections.abc import Callable, Mapping, Sequence

T = TypeVar("T")

_logger = get_logger(__name__)


class _Unset:
    """
    A marker class to indicate that a value is unset and thus should
    not be considered. This is different to None.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __repr__(self) -> str:
        return "<UNSET>"

    def __bool__(self) -> Literal[False]:
        return False

    def __copy__(self):
        return UNSET

    def __deepcopy__(self, memo: dict[int, Any]):
        return UNSET


UNSET = _Unset()


def is_unset(value: Any) -> bool:
    """
    Returns whether the given value is an instance of Unset.
    """
    return value is UNSET


def is_set_and_valid(value: Any) -> bool:
    return not is_unset(value) and value is not None


def is_set_and_present(value: T | None) -> TypeGuard[T]:
    return is_set_and_valid(value)


def unwrap(value: T | None, error_message: str = "unexpected None when unwrapping value") -> T:
    """
    Will unwrap the given value or raise a ValueError if it is None

    :param value: the optional value to unwrap
    :param error_message: the error message when failing to unwrap
    :return: the value or a ValueError if it is None
    """
    if value is None:
        raise ValueError(error_message)
    else:
        return value


def expect_type(value: Any, expected_type: type[T]) -> T:
    if isinstance(value, expected_type):
        return value
    else:
        raise ValueError(f"unexpected value of type '{type(value)}' while '{expected_type}' was expected")


@dataclass
class Change(Generic[T]):
    from_value: T | None
    to_value: T | None


def is_different_ignoring_order(value: Any, other_value: Any) -> bool:
    """
    Checks whether two values are considered to be equal.
    Note: two lists are considered to be equal if they contain the same elements,
    regardless or the order.
    """
    if isinstance(value, list) and isinstance(other_value, list):
        return sorted(value) != sorted(other_value)
    elif isinstance(value, dict) and isinstance(other_value, dict):
        for key, item_value in value.items():
            other_item_value = other_value.get(key, None)
            if isinstance(item_value, list):
                if sorted(item_value) != sorted(other_item_value):
                    return True
            else:
                if item_value != other_item_value:
                    return True

        return False

    return value != other_value


def patch_to_other(value: Any, other_value: Any) -> tuple[bool, Any]:
    if isinstance(value, dict):
        if len(other_value) == 0:
            if len(value) > 0:
                return True, value
            else:
                return False, None
        else:
            diff = dict(value)

            for k, v in other_value.items():
                if diff.get(k) == v:
                    diff.pop(k)

            if len(diff) == 0:
                return False, None
            else:
                return True, diff
    elif isinstance(value, list):
        sorted_value_list = sorted(value)

        if other_value is None:
            return True, sorted_value_list

        sorted_other_list = sorted(other_value)

        if sorted_value_list != sorted_other_list:
            diff = _diff_list(sorted_value_list, sorted_other_list)  # type: ignore
            if len(diff) > 0:
                return True, diff
            else:
                return False, diff
    else:
        if value != other_value:
            return True, value

    # values are not different, no patch generated
    return False, None


def _diff_list(list1: list[T], list2: list[T]) -> list[T]:
    s = set(list2)
    return [x for x in list1 if x not in s]


def write_patch_object_as_json(
    diff_object: dict[str, Any], printer: IndentingPrinter, close_object: bool = True
) -> None:
    if close_object is True and len(diff_object) == 0:
        printer.println(",")
        return

    printer.println(" {")
    printer.level_up()

    for key, value in sorted(diff_object.items()):
        if is_unset(value):
            _logger.warning("key '%s' defined in default configuration not present in config, skipping", key)
            continue

        if isinstance(value, list):
            printer.println(f"{key}+: [")
            printer.level_up()
            num_items = len(value)
            for index, item in enumerate(value):
                if index < num_items - 1:
                    printer.println(f"{json.dumps(item, ensure_ascii=False)},")
                else:
                    printer.println(f"{json.dumps(item, ensure_ascii=False)}")
            printer.level_down()
            printer.println("],")
        elif isinstance(value, dict):
            printer.println(f"{key}+: {{")
            printer.level_up()
            num_items = len(value)
            for index, (k, v) in enumerate(value.items()):
                if index < num_items - 1:
                    printer.println(f"{k}: {json.dumps(v, ensure_ascii=False)},")
                else:
                    printer.println(f"{k}: {json.dumps(v, ensure_ascii=False)}")
            printer.level_down()
            printer.println("},")
        else:
            printer.println(f"{key}: {json.dumps(value, ensure_ascii=False)},")

    if close_object is True:
        printer.level_down()
        printer.println("},")


def associate_by_key(input_list: Sequence[T], key_func: Callable[[T], str]) -> dict[str, T]:
    result = {}
    for item in input_list:
        key = key_func(item)

        if key in result:
            raise RuntimeError(f"duplicate item found with key '{key}'")

        result[key] = item

    return result


def multi_associate_by_key(input_list: Sequence[T], key_func: Callable[[T], list[str]]) -> dict[str, T]:
    result = {}
    for item in input_list:
        keys = key_func(item)

        for key in keys:
            if key in result:
                raise RuntimeError(f"duplicate item found with key '{key}'")

            result[key] = item

    return result


class LogLevel(Enum):
    GLOBAL = 0
    INFO = 1
    WARN = 2
    ERROR = 3


class IndentingPrinter:
    def __init__(
        self,
        output: TextIO | Console,
        initial_offset: int = 0,
        spaces_per_level: int = 2,
        log_level: LogLevel = LogLevel.GLOBAL,
        output_for_github: bool = False,
    ):
        if isinstance(output, Console):
            self._console = output
            self._writer = self._console.file
        else:
            no_color = output_for_github
            width = (
                110 if output_for_github else 9999
            )  # set a ridiculous high width for the normal case, otherwise rich will try to wrap lines
            self._console = Console(file=output, no_color=no_color, width=width)
            self._writer = output

        self._initial_offset = " " * initial_offset
        self._level = 0
        self._spaces_per_level = spaces_per_level
        self._indented_line = False
        self._log_level = log_level
        self._output_for_github = output_for_github

    @property
    def spaces_per_level(self) -> int:
        return self._spaces_per_level

    @property
    def console(self) -> Console:
        return self._console

    @property
    def writer(self) -> TextIO:
        return self._writer  # type: ignore

    @property
    def current_indentation(self) -> str:
        return self._initial_offset + " " * (self._level * self._spaces_per_level)

    def print(self, text: str = "", highlight: bool = False) -> None:
        lines = text.splitlines(keepends=True)
        if len(lines) > 0:
            for line in lines:
                self._print_indentation()

                if line.endswith("\n"):
                    self._console.print(line[:-1], end="", highlight=highlight)
                    self.print_line_break()
                else:
                    self._console.print(line, end="", highlight=highlight)

    def println(self, text: str = "", highlight: bool = False) -> None:
        self.print(text, highlight=highlight)
        self.print_line_break()

    def print_line_break(self) -> None:
        self._console.print("")
        self._indented_line = False

    def _print_indentation(self) -> None:
        if not self._indented_line:
            self._console.print(self.current_indentation, end="")
            self._indented_line = True

    def _is_logging_enabled(self, level: int, global_fn: Callable[[], bool]) -> bool:
        match self._log_level:
            case LogLevel.GLOBAL:
                return global_fn()

            case _:
                return self._log_level.value <= level

    def print_info(self, msg: str) -> None:
        if self._is_logging_enabled(1, is_info_enabled):
            _print_message(msg, "green", "Info", self._console, "+" if self._output_for_github else None)

    def is_info_enabled(self) -> bool:
        return self._is_logging_enabled(1, is_info_enabled)

    def print_warn(self, msg: str) -> None:
        if self._is_logging_enabled(2, lambda: True):
            _print_message(msg, "yellow", "Warning", self._console, "!" if self._output_for_github else None)

    def print_error(self, msg: str) -> None:
        if self._is_logging_enabled(3, lambda: True):
            _print_message(msg, "red", "Error", self._console, "-" if self._output_for_github else None)

    def level_up(self) -> None:
        self._level += 1

    def level_down(self) -> None:
        if self._level == 0:
            raise RuntimeError("tried to call level_down on level 0")

        self._level -= 1


async def run_command(cmd: str, *args: str, **kwargs) -> tuple[int, str, str]:
    import asyncio

    process = await asyncio.create_subprocess_exec(
        cmd, *args, **kwargs, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )

    stdout, stderr = await process.communicate()
    return process.returncode, stdout.decode("utf-8"), stderr.decode("utf-8")  # type: ignore


def jsonnet_evaluate_file(file: str) -> dict[str, Any]:
    import rjsonnet

    _logger.trace("evaluating jsonnet file '%s'", file)

    try:
        return json.loads(rjsonnet.evaluate_file(file))
    except Exception as ex:
        raise RuntimeError(f"failed to evaluate jsonnet file: {ex!s}") from ex


def jsonnet_evaluate_snippet(snippet: str) -> dict[str, Any]:
    import rjsonnet

    _logger.trace("evaluating jsonnet snippet '%s'", snippet)

    try:
        return json.loads(rjsonnet.evaluate_snippet("", snippet))
    except Exception as ex:
        raise RuntimeError(f"failed to evaluate snippet: {ex!s}") from ex


def get_or_default(namespace: Namespace, key: str, default: T) -> T:
    if namespace.__contains__(key):
        return namespace.__getattribute__(key)
    else:
        return default


def snake_to_camel_case(string: str) -> str:
    result = re.sub(r"[_\-]+", " ", string).title().replace(" ", "")
    return result[0].lower() + result[1:]


def snake_to_normal_case(string: str) -> str:
    result = re.sub(r"[_\-]+", " ", string).title()
    return result[0] + result[1:]


def camel_to_snake_case(string: str) -> str:
    string = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", string)
    string = re.sub("__([A-Z])", r"_\1", string)
    string = re.sub("([a-z0-9])([A-Z])", r"\1_\2", string)
    return string.lower()


def parse_template_url(url: str) -> tuple[str, str, str]:
    parts = urlparse(url)

    if parts.netloc != "github.com":
        raise ValueError(f"only github.com is supported for template urls: {parts.netloc}")

    repo_url = f"{parts.scheme}://{parts.netloc}{parts.path}"

    matcher = re.match(r"([^@]+)(@(.*))?", parts.fragment)
    if matcher is None:
        raise ValueError(f"failed to parse file from template url: {url}")

    file = matcher.group(1)
    ref = matcher.group(3)

    if ref is None:
        raise ValueError(f"failed to parse ref from template url: {url}")

    return repo_url, file, ref


def parse_github_url(url: str) -> tuple[str, str]:
    pattern = re.compile(r"https://github.com/([A-Za-z0-9_.\-]+)/([A-Za-z0-9_.\-]+)")
    m = pattern.match(url)
    if m is None:
        raise ValueError(f"unexpected GitHub url '{url}'")
    else:
        return m.group(1), m.group(2)


def is_ghsa_repo(repo_name: str) -> bool:
    """
    Returns True if the given repo_name is considered to be a repo created for a GitHub security advisory.

    Pattern: <regular repo name>-ghsa-xxxx-xxxx-xxxx
    """
    return re.match(".*-ghsa(-[23456789cfghjmpqrvwx]{4}){3}", repo_name) is not None


def strip_trailing_commas(lines: list[str]) -> list[str]:
    return [line.rstrip(",") for line in lines]


def sort_jsonnet(lines: list[str]) -> list[str]:
    ast: list[tuple[str, Any]] = []
    object_stack = [ast]

    for line in lines:
        trimmed_line = line.rstrip().rstrip(",")
        if trimmed_line.endswith(("{", "[")):
            current_node: list[tuple[str, Any]] = []
            object_stack[-1].append((line, current_node))
            object_stack.append(current_node)
        elif (trimmed_line.endswith("}") and "{" not in trimmed_line) or (
            trimmed_line.endswith("]") and "[" not in trimmed_line
        ):
            object_stack[-1].append((line, None))
            object_stack.pop()
        else:
            object_stack[-1].append((line, None))

    for node in ast:
        _sort_node(node)

    result: list[str] = []
    for node in ast:
        print_node(node, result)
    return result


def _sort_node(node):
    line, context = node

    if context is not None:
        last = context.pop()
        context.sort()
        context.append(last)
        for next_node in context:
            _sort_node(next_node)


def print_node(node, result):
    line, context = node

    result.append(line)
    if context is not None:
        for next_node in context:
            print_node(next_node, result)


def get_approval() -> bool:
    answer = input()
    if answer != "yes" and answer != "y":
        return False
    else:
        return True


class PrettyFormatter:
    def __init__(self, spaces_per_level: int = 2, key_align: int = -1):
        self.types: dict[str, Any] = {}
        self.htchar = " " * spaces_per_level
        self.lfchar = "\n"
        self.key_align = key_align
        self.indent = 0
        self.set_formatter(object, self.__class__._format_object)
        self.set_formatter(dict, self.__class__._format_dict)
        self.set_formatter(list, self.__class__._format_list)
        self.set_formatter(tuple, self.__class__._format_tuple)

    def set_formatter(self, obj, callback):
        self.types[obj] = callback

    def format(self, value, **args):
        for key in args:
            setattr(self, key, args[key])
        formatter = self.types[type(value) if type(value) in self.types else object]
        return formatter(self, value, self.indent)

    def _format_object(self, value, indent):
        if isinstance(value, str):
            if "\n" in value:
                return '"""\n' + value + '"""'

        return json.dumps(value, ensure_ascii=False)

    def _format_dict(self, value, indent):
        if len(value) == 0:
            return "{}"

        key_align = max(len(repr(x)) for x in value) + 1 if self.key_align == -1 else self.key_align

        items = [
            self.lfchar
            + self.htchar * (indent + 1)
            + repr(key).ljust(key_align)
            + ": "
            + (self.types[type(value[key]) if type(value[key]) in self.types else object])(self, value[key], indent + 1)
            for key in sorted(value)
        ]
        return "{%s}" % (",".join(items) + self.lfchar + self.htchar * indent)

    def _format_list(self, value, indent):
        items = [
            self.lfchar
            + self.htchar * (indent + 1)
            + (self.types[type(item) if type(item) in self.types else object])(self, item, indent + 1)
            for item in value
        ]
        if len(items) == 0:
            return "[]"
        else:
            return "[%s]" % (",".join(items) + self.lfchar + self.htchar * indent)

    def _format_tuple(self, value, indent):
        items = [
            self.lfchar
            + self.htchar * (indent + 1)
            + (self.types[type(item) if type(item) in self.types else object])(self, item, indent + 1)
            for item in value
        ]
        return "(%s)" % (",".join(items) + self.lfchar + self.htchar * indent)


def query_json(expr: str, data: Mapping[str, Any]) -> Any:
    """
    Evaluates a jsonata expression on the given dictionary.
    """
    from jsonata import Jsonata  # type: ignore

    return Jsonata.jsonata(expr).evaluate(data)


def deep_merge_dict(source: dict[str, Any], destination: dict[str, Any]):
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            deep_merge_dict(value, node)
        else:
            destination[key] = value

    return destination


def format_date_for_csv(iso_date_str: str) -> str:
    if iso_date_str is None:
        return ""
    date_obj = datetime.fromisoformat(iso_date_str)
    return date_obj.strftime("%d/%m/%Y %H:%M:%S")


def debug_times(category: str):
    import asyncio
    import functools

    def decorator_timed(func):
        if asyncio.iscoroutinefunction(func):

            @functools.wraps(func)
            async def wrapper_timed(*args, **kwargs):
                start = datetime.now()
                _logger.debug(f"{category}: starting ...")
                value = await func(*args, **kwargs)
                end = datetime.now()
                _logger.debug(f"{category}: complete after {(end - start).total_seconds()}s")
                return value

            return wrapper_timed
        else:

            @functools.wraps(func)
            def wrapper_timed(*args, **kwargs):
                start = datetime.now()
                _logger.debug(f"{category}: starting ...")
                value = func(*args, **kwargs)
                end = datetime.now()
                _logger.debug(f"{category}: complete after {(end - start).total_seconds()}s")
                return value

            return wrapper_timed

    return decorator_timed


def render_chevron(content: str, context: dict[str, Any]) -> str:
    import chevron

    # need to escape ${{ sequences as used at GitHub
    escaped_content = content.replace("${{", "$\\{\\{")
    output = chevron.render(escaped_content, context)
    return output.replace("$\\{\\{", "${{")



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import filecmp
from typing import TYPE_CHECKING

from aiofiles import open

from otterdog.providers.github import GitHubProvider
from otterdog.utils import get_approval

from . import Operation
from .local_plan import LocalPlanOperation

if TYPE_CHECKING:
    from otterdog.config import OrganizationConfig
    from otterdog.providers.github.rest import RestApi


class OpenPullRequestOperation(Operation):
    """
    Creates a pull request for local configuration changes of an organization in its meta-data repository.
    """

    def __init__(self, branch: str, title: str, author: str | None):
        super().__init__()
        self._branch = branch
        self._title = title
        self._author = author

    @property
    def branch(self) -> str:
        return self._branch

    @property
    def title(self) -> str:
        return self._title

    @property
    def author(self) -> str | None:
        return self._author

    def pre_execute(self) -> None:
        self.printer.println("Open PR for local configuration changes:")

    async def execute(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        await jsonnet_config.init_template()

        self._print_project_header(org_config, org_index, org_count)

        org_file_name = jsonnet_config.org_config_file
        if not await self.check_config_file_exists(org_file_name):
            return 1

        try:
            credentials = self.get_credentials(org_config, only_token=True)
        except RuntimeError as ex:
            self.printer.print_error(f"invalid credentials\n{ex!s}")
            return 1

        self.printer.level_up()

        try:
            async with open(org_file_name) as file:
                local_configuration = await file.read()

            async with GitHubProvider(credentials) as provider:
                rest_api = provider.rest_api

                try:
                    default_branch = await rest_api.repo.get_default_branch(
                        org_config.github_id, org_config.config_repo
                    )

                    current_definition = await provider.get_content(
                        github_id,
                        org_config.config_repo,
                        f"otterdog/{github_id}.jsonnet",
                        default_branch,
                    )

                    current_config_file = org_config.jsonnet_config.org_config_file + "-BASE"
                    async with open(current_config_file, "w") as file:
                        await file.write(current_definition)

                    if filecmp.cmp(current_config_file, org_config.jsonnet_config.org_config_file):
                        self.printer.println("no local changes, no PR has been opened")
                        return 0

                    self.printer.println("The following changes compared to the current configuration exist locally:")
                    self.printer.println()
                    self.printer.level_up()

                    try:
                        operation = LocalPlanOperation("-BASE", "*", False, False, "")
                        operation.init(self.config, self.printer)
                        valid_config = await operation.generate_diff(org_config)
                    finally:
                        self.printer.level_down()

                    if valid_config != 0:
                        self.printer.println("the local configuration contains validation error")
                        return 1

                    self.printer.println()
                    self.printer.println(
                        "Do you want to open a PR with these changes? "
                        "(Only 'yes' or 'y' will be accepted as approval)\n"
                    )

                    self.printer.print("[bold]Enter a value:[/] ")
                    if not get_approval():
                        self.printer.println("\nOpen PR cancelled.")
                        return 1

                    pr_number = await self._create_pull_request(
                        rest_api,
                        org_config,
                        default_branch,
                        local_configuration,
                    )

                    self.printer.println(
                        f"created pull request #{pr_number} at "
                        f"https://github.com/{org_config.github_id}/{org_config.config_repo}/pull/{pr_number}"
                    )
                except RuntimeError as e:
                    self.printer.print_error(
                        "failed to open pull request in repo "
                        f"'{org_config.github_id}/{org_config.config_repo}': {e!s}"
                    )
                    return 1

            return 0
        finally:
            self.printer.level_down()

    async def _create_pull_request(
        self,
        rest_api: RestApi,
        org_config: OrganizationConfig,
        default_branch: str,
        local_configuration: str,
    ) -> str:
        default_branch_data = await rest_api.reference.get_branch_reference(
            org_config.github_id,
            org_config.config_repo,
            default_branch,
        )
        default_branch_sha = default_branch_data["object"]["sha"]
        branch_name = f"otterdog/{self.branch}"

        await rest_api.reference.create_reference(
            org_config.github_id,
            org_config.config_repo,
            branch_name,
            default_branch_sha,
        )

        await rest_api.content.update_content(
            org_config.github_id,
            org_config.config_repo,
            f"otterdog/{org_config.github_id}.jsonnet",
            local_configuration,
            branch_name,
        )

        if self.author is not None:
            body = f"This PR has been created automatically on behalf of @{self.author} using the otterdog cli."
        else:
            body = "This PR has been created automatically using the otterdog cli."

        pull_request_data = await rest_api.pull_request.create_pull_request(
            org_config.github_id,
            org_config.config_repo,
            self.title,
            branch_name,
            default_branch,
            body,
        )

        return pull_request_data["number"]



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import json
import os
import unittest
from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import Any
from unittest.mock import MagicMock

from otterdog.models import ModelObject
from otterdog.utils import jsonnet_evaluate_snippet

_EXAMPLE_CONFIG = "./examples/template/otterdog-defaults.libsonnet"


class ModelTest(ABC, unittest.IsolatedAsyncioTestCase):
    @property
    def org_id(self) -> str:
        return "OtterdogTest"

    @property
    @abstractmethod
    def template_function(self) -> str:
        pass

    @abstractmethod
    def create_model(self, data: Mapping[str, Any]) -> ModelObject:
        pass

    @property
    @abstractmethod
    def model_data(self):
        pass

    @property
    @abstractmethod
    def provider_data(self):
        pass

    @property
    def provider(self):
        async def get_actor_node_ids(actors):
            return [f"id_{actor[1:]}" for actor in actors]

        async def get_app_node_ids(app_names):
            return {app: f"id_{app}" for app in app_names}

        async def get_actor_ids_with_type(actors):
            result = []
            for actor in actors:
                if "/" in actor:
                    result.append(("Team", (f"id_{actor[1:]}", f"id_{actor[1:]}")))
                else:
                    result.append(("User", (f"id_{actor[1:]}", f"id_{actor[1:]}")))
            return result

        provider = MagicMock()

        provider.get_actor_node_ids = MagicMock(side_effect=get_actor_node_ids)
        provider.get_app_node_ids = MagicMock(side_effect=get_app_node_ids)
        provider.get_actor_ids_with_type = MagicMock(side_effect=get_actor_ids_with_type)

        return provider

    @staticmethod
    def load_json_resource(file: str) -> dict[str, Any]:
        filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), f"resources/{file}")
        with open(filename) as fp:
            return json.load(fp)

    def test_example_config(self) -> None:
        snippet = f"(import '{_EXAMPLE_CONFIG}').{self.template_function}('default')"
        json_data = jsonnet_evaluate_snippet(snippet)

        model_object = self.create_model(json_data)
        assert model_object is not None
        if model_object.is_keyed():
            assert model_object.get_key_value() == "default"



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from . import CacheStrategy

if TYPE_CHECKING:
    from typing import Any

    from aiohttp_client_cache import CacheBackend

_AIOHTTP_CACHE_DIR = ".cache/async_http"


def file_cache(cache_dir: str = _AIOHTTP_CACHE_DIR) -> CacheStrategy:
    return _FileCache(cache_dir)


class _FileCache(CacheStrategy):
    def __init__(self, cache_dir: str):
        self._cache_dir = cache_dir

    def get_cache_backend(self) -> CacheBackend:
        from aiohttp_client_cache.backends import FileBackend

        return FileBackend(
            cache_name=self._cache_dir,
            use_temp=False,
        )

    def is_external(self) -> bool:
        return False

    def get_request_parameters(self) -> dict[str, Any]:
        return {"refresh": True}

    def __str__(self):
        return f"file-cache('{self._cache_dir}')"



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from otterdog.providers.github import GitHubProvider

from . import Operation

if TYPE_CHECKING:
    from otterdog.config import OrganizationConfig


class WebLoginOperation(Operation):
    """
    Opens up a logged in browser for an organization.
    """

    def __init__(self):
        super().__init__()

    def pre_execute(self) -> None:
        pass

    async def execute(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> int:
        github_id = org_config.github_id

        self._print_project_header(org_config, org_index, org_count)
        self.printer.level_up()

        try:
            try:
                credentials = self.get_credentials(org_config)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{e!s}")
                return 1

            async with GitHubProvider(credentials) as provider:
                await provider.web_client.open_browser_with_logged_in_user(github_id)

            return 0
        finally:
            self.printer.level_down()



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import abc
import dataclasses
from typing import TYPE_CHECKING, Any, TypeVar

from jsonbender import F, Forall, OptionalS, S  # type: ignore

from otterdog.models import (
    FailureType,
    LivePatch,
    LivePatchType,
    ModelObject,
    ValidationContext,
)
from otterdog.utils import UNSET, is_set_and_valid, unwrap

if TYPE_CHECKING:
    from otterdog.jsonnet import JsonnetConfig
    from otterdog.providers.github import GitHubProvider

TT = TypeVar("TT", bound="Team")


@dataclasses.dataclass
class Team(ModelObject, abc.ABC):
    """
    Represents a Team.
    """

    id: int = dataclasses.field(metadata={"external_only": True})
    slug: str = dataclasses.field(metadata={"external_only": True})
    name: str = dataclasses.field(metadata={"key": True})
    description: str | None
    privacy: str
    notifications: bool
    members: list[str]
    skip_members: bool = dataclasses.field(metadata={"model_only": True}, default=False)
    skip_non_organization_members: bool = dataclasses.field(metadata={"model_only": True}, default=False)

    @property
    def model_object_name(self) -> str:
        return "team"

    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> str | None:
        return f"orgs.{jsonnet_config.create_org_team}"

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        if field.name == "members":
            return not self.skip_members

        return True

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        # execute custom validation rules if present
        self.execute_custom_validation_if_present(context, "validate-team.py")

        if (
            context.exclude_teams_pattern is not None
            and context.exclude_teams_pattern.match(self.name)
            and self.name not in context.default_team_names
        ):
            context.add_failure(
                FailureType.ERROR,
                f"{self.get_model_header(parent_object)} has 'name' of value '{self.name}', "
                f"which is not allowed due to exclusion pattern '{context.exclude_teams_pattern.pattern}'.",
            )

        if is_set_and_valid(self.privacy):
            if self.privacy not in {"secret", "visible"}:
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'privacy' of value '{self.privacy}', "
                    f"while only values ('secret' | 'closed') are allowed.",
                )

        if self.skip_members is True and is_set_and_valid(self.members) and len(self.members) > 0:
            context.add_failure(
                FailureType.ERROR,
                f"{self.get_model_header(parent_object)} has 'skip_members' enabled, "
                f"but 'members' is set to {self.members}.",
            )

        if is_set_and_valid(self.members) and self.skip_non_organization_members is True:
            for member in self.members:
                if member not in context.org_members:
                    context.add_failure(
                        FailureType.ERROR,
                        f"{self.get_model_header(parent_object)} has 'skip_non_organization_members' enabled, "
                        f"but 'members' contains user '{member}' who is not an organization member.",
                    )

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = super().get_mapping_from_provider(org_id, data)

        def transform_notification_setting(x: str | None):
            if x is None:
                return UNSET
            elif x == "notifications_enabled":
                return True
            else:
                return False

        def transform_team_members(member):
            return member["login"]

        mapping.update(
            {
                "privacy": OptionalS("privacy") >> F(lambda x: "visible" if x == "closed" else x),
                "notifications": OptionalS("notification_setting") >> F(transform_notification_setting),
                "members": OptionalS("members", default=[]) >> Forall(transform_team_members),
            }
        )
        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        mapping = await super().get_mapping_to_provider(org_id, data, provider)

        if "privacy" in data:
            mapping["privacy"] = S("privacy") >> F(lambda x: "closed" if x == "visible" else x)

        if "notifications" in data:
            mapping["notification_setting"] = S("notifications") >> F(
                lambda x: "notifications_enabled" if x is True else "notifications_disabled"
            )
            mapping.pop("notifications")

        return mapping

    @classmethod
    async def apply_live_patch(
        cls,
        patch: LivePatch[Team],
        org_id: str,
        provider: GitHubProvider,
    ) -> None:
        match patch.patch_type:
            case LivePatchType.ADD:
                expected_object = unwrap(patch.expected_object)
                await provider.add_org_team(
                    org_id,
                    expected_object.name,
                    await expected_object.to_provider_data(org_id, provider),
                )

            case LivePatchType.REMOVE:
                await provider.delete_org_team(org_id, unwrap(patch.current_object).slug)

            case LivePatchType.CHANGE:
                await provider.update_org_team(
                    org_id,
                    unwrap(patch.current_object).slug,
                    await cls.changes_to_provider(org_id, unwrap(patch.changes), provider),
                )



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
import os
from abc import ABC, abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, Any, Generic, Protocol, Self, TypeVar, cast, final

from jsonbender import OptionalS, S, bend  # type: ignore

from otterdog.utils import (
    UNSET,
    Change,
    IndentingPrinter,
    T,
    associate_by_key,
    is_different_ignoring_order,
    is_set_and_valid,
    is_unset,
    multi_associate_by_key,
    patch_to_other,
    unwrap,
    write_patch_object_as_json,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Mapping, Sequence
    from re import Pattern

    from otterdog.config import SecretResolver
    from otterdog.jsonnet import JsonnetConfig
    from otterdog.providers.github import GitHubProvider

MT = TypeVar("MT", bound="ModelObject")
EMT = TypeVar("EMT", bound="EmbeddedModelObject")


class FailureType(Enum):
    INFO = 1
    WARNING = 2
    ERROR = 3


@dataclasses.dataclass
class ValidationContext:
    root_object: Any
    secret_resolver: SecretResolver
    template_dir: str
    org_members: set[str]
    default_team_names: set[str]
    exclude_teams_pattern: Pattern | None
    validation_failures: list[tuple[FailureType, str]] = dataclasses.field(default_factory=list)

    def add_failure(self, failure_type: FailureType, message: str):
        self.validation_failures.append((failure_type, message))

    def property_equals(self, model_object, key, value):
        current_value = model_object.__getattribute__(key)
        if current_value != value:
            self.add_failure(
                FailureType.ERROR,
                f"{model_object.get_model_header()} has '{key}' set to '{current_value}' but '{value}' is required.",
            )


class LivePatchType(Enum):
    ADD = 1
    REMOVE = 2
    CHANGE = 3


class LivePatchApplyFn(Protocol[MT]):
    async def __call__(self, patch: LivePatch[MT], org_id: str, provider: GitHubProvider) -> None: ...


@dataclasses.dataclass(frozen=True)
class LivePatch(Generic[MT]):
    patch_type: LivePatchType
    expected_object: MT | None
    current_object: MT | None
    changes: dict[str, Change] | None
    parent_object: ModelObject | None
    forced_update: bool
    fn: LivePatchApplyFn
    changes_object_to_readonly: bool = False

    @classmethod
    def of_addition(cls, expected_object: MT, parent_object: ModelObject | None, fn: LivePatchApplyFn[MT]) -> LivePatch:
        return LivePatch(LivePatchType.ADD, expected_object, None, None, parent_object, False, fn)

    @classmethod
    def of_deletion(cls, current_object: MT, parent_object: ModelObject | None, fn: LivePatchApplyFn[MT]) -> LivePatch:
        return LivePatch(LivePatchType.REMOVE, None, current_object, None, parent_object, False, fn)

    @classmethod
    def of_changes(
        cls,
        expected_object: MT,
        current_object: MT,
        changes: dict[str, Change],
        parent_object: ModelObject | None,
        forced_update: bool,
        fn: LivePatchApplyFn[MT],
        changes_object_to_readonly: bool = False,
    ) -> LivePatch:
        return LivePatch(
            LivePatchType.CHANGE,
            expected_object,
            current_object,
            changes,
            parent_object,
            forced_update,
            fn,
            changes_object_to_readonly,
        )

    def requires_web_ui(self) -> bool:
        match self.patch_type:
            case LivePatchType.ADD:
                return unwrap(self.expected_object).requires_web_ui()

            case LivePatchType.REMOVE:
                return False

            case LivePatchType.CHANGE:
                return unwrap(self.expected_object).changes_require_web_ui(unwrap(self.changes))

    def requires_secrets(self) -> bool:
        match self.patch_type:
            case LivePatchType.ADD:
                return unwrap(self.expected_object).contains_secrets()

            case LivePatchType.REMOVE:
                return False

            case LivePatchType.CHANGE:
                return unwrap(self.expected_object).contains_secrets()

    async def apply(self, org_id: str, provider: GitHubProvider) -> None:
        await self.fn(self, org_id, provider)

    def __repr__(self) -> str:
        obj = self.expected_object if self.expected_object is not None else self.current_object
        return f"{self.patch_type.name} - {unwrap(obj).get_model_header(self.parent_object)}"


@dataclasses.dataclass
class PatchContext:
    org_id: str
    org_settings: ModelObject


@dataclasses.dataclass
class LivePatchContext:
    org_id: str
    repo_filter: str
    update_webhooks: bool
    update_secrets: bool
    update_filter: str
    current_org_settings: ModelObject | None
    expected_org_settings: ModelObject
    modified_org_settings: dict[str, Change] = dataclasses.field(default_factory=dict)
    modified_org_workflow_settings: dict[str, Change] = dataclasses.field(default_factory=dict)


class LivePatchHandler(Protocol):
    def __call__(self, patch: LivePatch) -> None: ...


@dataclasses.dataclass
class EmbeddedModelObject(ABC):
    """
    The abstract base class for embedded model objects.
    """

    @abstractmethod
    def validate(self, context: ValidationContext, parent_object: Any) -> None: ...

    def get_difference_from(self, other: Self) -> Change | None:
        if not isinstance(other, self.__class__):
            raise ValueError(f"'types do not match: {type(self)}' != '{type(other)}'")

        from_dict: dict[str, Any] = {}
        to_dict: dict[str, Any] = {}

        for key in self.keys(
            for_diff=True,
            for_patch=False,
            exclude_unset_keys=True,
        ):
            to_value = self.__getattribute__(key)
            from_value = other.__getattribute__(key)

            if is_unset(from_value):
                continue

            from_dict[key] = from_value
            to_dict[key] = to_value

        if is_different_ignoring_order(from_dict, to_dict):
            return Change(from_dict, to_dict)
        else:
            return None

    def get_patch_to(self, other: EmbeddedModelObject) -> dict[str, Any]:
        if not isinstance(other, self.__class__):
            raise ValueError(f"'types do not match: {type(self)}' != '{type(other)}'")

        patch_result = {}
        for key in self.keys(for_diff=False, for_patch=True, exclude_unset_keys=True):
            value = self.__getattribute__(key)
            other_value = other.__getattribute__(key)

            if is_unset(other_value):
                continue

            patch_needed, diff = patch_to_other(value, other_value)
            if patch_needed is True:
                patch_result[key] = diff

        return patch_result

    @abstractmethod
    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> str | None: ...

    def to_jsonnet(
        self,
        printer: IndentingPrinter,
        jsonnet_config: JsonnetConfig,
        context: PatchContext,
        extend: bool,
        default_object: EmbeddedModelObject,
    ) -> None:
        patch = self.get_patch_to(default_object)

        template_function = self.get_jsonnet_template_function(jsonnet_config, False)

        if extend is False:
            printer.print(f" {unwrap(template_function)}()")

        write_patch_object_as_json(patch, printer)

    @classmethod
    def all_fields(cls) -> list[dataclasses.Field]:
        return list(dataclasses.fields(cls))

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        return True

    def include_field_for_patch_computation(self, field: dataclasses.Field) -> bool:
        return self.include_field_for_diff_computation(field)

    def keys(
        self,
        for_diff: bool = False,
        for_patch: bool = False,
        exclude_unset_keys: bool = True,
    ) -> list[str]:
        result = []

        for field in self.all_fields():
            if for_diff is True and not self.include_field_for_diff_computation(field):
                continue

            if for_patch is True and not self.include_field_for_patch_computation(field):
                continue

            if exclude_unset_keys:
                value = self.__getattribute__(field.name)
                if not is_unset(value):
                    result.append(field.name)
            else:
                result.append(field.name)

        return result

    def to_model_dict(self, for_diff: bool = False) -> dict[str, Any]:
        result = {}

        for key in self.keys(for_diff=for_diff, for_patch=False, exclude_unset_keys=True):
            value = self.__getattribute__(key)
            result[key] = value

        return result

    @classmethod
    def from_model_data(cls: type[EMT], data: dict[str, Any]) -> EMT:
        mapping = cls.get_mapping_from_model()
        return cls(**bend(mapping, data))  # type: ignore

    @classmethod
    def get_mapping_from_model(cls) -> dict[str, Any]:
        return {k: OptionalS(k, default=UNSET) for k in (x.name for x in cls.all_fields())}

    @classmethod
    def from_provider_data(cls: type[EMT], org_id: str, data: dict[str, Any]) -> EMT:
        mapping = cls.get_mapping_from_provider(org_id, data)
        return cls(**bend(mapping, data))  # type: ignore

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        return {k: OptionalS(k, default=UNSET) for k in (x.name for x in cls.all_fields())}

    @classmethod
    async def dict_to_provider_data(cls, org_id: str, data: dict[str, Any], provider: GitHubProvider) -> dict[str, Any]:
        mapping = await cls.get_mapping_to_provider(org_id, data, provider)
        return bend(mapping, data)

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        return {field.name: S(field.name) for field in cls.all_fields() if not is_unset(data.get(field.name, UNSET))}


@dataclasses.dataclass
class ModelObject(ABC):
    """
    The abstract base class for any model object.
    """

    def __post_init__(self):
        """
        Assigns to all field which are UNSET their default value, if one is available.
        """
        for field in self.all_fields():
            value = self.__getattribute__(field.name)
            if is_unset(value):
                if field.default is not dataclasses.MISSING:
                    self.__setattr__(field.name, field.default)
                elif field.default_factory is not dataclasses.MISSING:
                    self.__setattr__(field.name, field.default_factory())

    @property
    @abstractmethod
    def model_object_name(self) -> str: ...

    def is_keyed(self) -> bool:
        """Indicates whether the ModelObject is keyed by a property"""
        return any(field.metadata.get("key", False) for field in self.all_fields())

    def get_key(self) -> str:
        """Returns the key property of this ModelObject if it keyed"""
        if not self.is_keyed():
            raise RuntimeError("model object is not keyed")

        return next(
            filter(
                lambda field: field.metadata.get("key", False) is True,
                self.all_fields(),
            )
        ).name

    def get_key_value(self) -> Any:
        """Returns the value of the key property"""
        return self.__getattribute__(self.get_key())

    def get_all_key_values(self) -> list[Any]:
        """Returns a list of all values by which this ModelObject is keyed"""
        return [self.get_key_value()]

    @abstractmethod
    def validate(self, context: ValidationContext, parent_object: Any) -> None: ...

    # noinspection PyMethodMayBeStatic
    def execute_custom_validation_if_present(self, context: ValidationContext, filename: str) -> None:
        validate_script = os.path.join(context.template_dir, filename)
        if os.path.exists(validate_script):
            with open(validate_script) as file:
                exec(file.read())

    def get_difference_from(self, other: Self) -> dict[str, Change[T]]:
        if not isinstance(other, self.__class__):
            raise ValueError(f"'types do not match: {type(self)}' != '{type(other)}'")

        diff_result: dict[str, Change[T]] = {}
        for key in self.keys(
            for_diff=True,
            for_patch=False,
            include_nested_models=False,
            exclude_unset_keys=True,
        ):
            to_value = self.__getattribute__(key)
            from_value = other.__getattribute__(key)

            if not other.is_key_valid_for_diff_computation(key, self):
                continue

            if is_unset(from_value):
                continue

            if isinstance(to_value, EmbeddedModelObject) and isinstance(from_value, EmbeddedModelObject):
                embedded_diff = to_value.get_difference_from(from_value)
                if embedded_diff is not None:
                    diff_result[key] = embedded_diff
                continue

            if isinstance(to_value, EmbeddedModelObject):
                to_value = to_value.to_model_dict()

            if isinstance(from_value, EmbeddedModelObject):
                from_value = from_value.to_model_dict()

            # TODO: improve diff computation for embedded model objects
            #       right now a simple dict comparison is being made
            #       but that does not take UNSET values into account
            if is_different_ignoring_order(to_value, from_value):
                diff_result[key] = Change(from_value, to_value)

        return diff_result

    def get_patch_to(self, other: ModelObject) -> dict[str, Any]:
        if not isinstance(other, self.__class__):
            raise ValueError(f"'types do not match: {type(self)}' != '{type(other)}'")

        patch_result = {}
        for key in self.keys(
            for_diff=False,
            for_patch=True,
            include_nested_models=False,
            exclude_unset_keys=True,
        ):
            value = self.__getattribute__(key)
            other_value = other.__getattribute__(key)

            if is_unset(other_value):
                continue

            patch_needed, diff = patch_to_other(value, other_value)
            if patch_needed is True:
                patch_result[key] = diff

        return patch_result

    @classmethod
    def all_fields(cls) -> list[dataclasses.Field]:
        return list(dataclasses.fields(cls))

    @classmethod
    def model_fields(cls) -> list[dataclasses.Field]:
        return [field for field in dataclasses.fields(cls) if not cls.is_external_only(field)]

    @classmethod
    def model_only_fields(cls) -> list[dataclasses.Field]:
        return [field for field in dataclasses.fields(cls) if cls.is_model_only(field)]

    @classmethod
    def provider_fields(cls) -> list[dataclasses.Field]:
        return [
            field
            for field in dataclasses.fields(cls)
            if not cls.is_external_only(field)
            and not cls.is_model_only(field)
            and not cls.is_read_only(field)
            and not cls.is_nested_model(field)
        ]

    @classmethod
    def _get_field(cls, key: str) -> dataclasses.Field:
        for field in dataclasses.fields(cls):
            if field.name == key:
                return field

        raise ValueError(f"unknown key {key}")

    @staticmethod
    def is_external_only(field: dataclasses.Field) -> bool:
        return field.metadata.get("external_only", False) is True

    @staticmethod
    def is_read_only(field: dataclasses.Field) -> bool:
        return field.metadata.get("read_only", False) is True

    @classmethod
    def is_read_only_key(cls, key: str) -> bool:
        return cls.is_read_only(cls._get_field(key))

    @classmethod
    def is_nested_model_key(cls, key: str) -> bool:
        return cls.is_nested_model(cls._get_field(key))

    @classmethod
    def is_embedded_model_key(cls, key: str) -> bool:
        return cls.is_embedded_model(cls._get_field(key))

    @staticmethod
    def is_model_only(field: dataclasses.Field) -> bool:
        return field.metadata.get("model_only", False) is True

    @staticmethod
    def is_nested_model(field: dataclasses.Field) -> bool:
        return field.metadata.get("nested_model", False) is True

    @staticmethod
    def is_embedded_model(field: dataclasses.Field) -> bool:
        return field.metadata.get("embedded_model", False) is True

    def get_model_objects(self) -> Iterator[tuple[ModelObject, ModelObject]]:
        yield from []

    def get_model_header(self, parent_object: ModelObject | None = None) -> str:
        header = f"[bold]{self.model_object_name}[/]"

        if self.is_keyed():
            key = self.get_key()
            header = header + f'\\[{key}="[bold]{self.get_key_value()}[/]"'

            if isinstance(parent_object, ModelObject) and parent_object.is_keyed():
                header = header + f", {parent_object.model_object_name}=" + f"[bold]{parent_object.get_key_value()}[/]"

            header = header + "]"
        elif isinstance(parent_object, ModelObject) and parent_object.is_keyed():
            header = header + "\\["
            header = header + f"{parent_object.model_object_name}=" + f"[bold]{parent_object.get_key_value()}[/]"
            header = header + "]"

        return header

    @classmethod
    @final
    def from_model_data(cls: type[MT], data: Mapping[str, Any]) -> MT:
        mapping = cls.get_mapping_from_model()
        return cls(**bend(mapping, data))  # type: ignore

    @classmethod
    def get_mapping_from_model(cls) -> dict[str, Any]:
        return {k: OptionalS(k, default=UNSET) for k in (x.name for x in cls.all_fields())}

    @classmethod
    @final
    def from_provider_data(cls: type[MT], org_id: str, data: dict[str, Any]) -> MT:
        mapping = cls.get_mapping_from_provider(org_id, data)
        return cls(**bend(mapping, data))  # type: ignore

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        return {k: OptionalS(k, default=UNSET) for k in (x.name for x in cls.all_fields())}

    async def to_provider_data(self, org_id: str, provider: GitHubProvider) -> dict[str, Any]:
        return await self.dict_to_provider_data(org_id, self.to_model_dict(), provider)

    @classmethod
    async def changes_to_provider(
        cls, org_id: str, data: dict[str, Change[Any]], provider: GitHubProvider
    ) -> dict[str, Any]:
        return await cls.dict_to_provider_data(org_id, {key: change.to_value for key, change in data.items()}, provider)

    @classmethod
    async def dict_to_provider_data(cls, org_id: str, data: dict[str, Any], provider: GitHubProvider) -> dict[str, Any]:
        mapping = await cls.get_mapping_to_provider(org_id, data, provider)
        return bend(mapping, data)

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        return {
            field.name: S(field.name) for field in cls.provider_fields() if not is_unset(data.get(field.name, UNSET))
        }

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        return True

    def is_key_valid_for_diff_computation(self, key: str, expected_object: Self) -> bool:
        return True

    def include_field_for_patch_computation(self, field: dataclasses.Field) -> bool:
        return self.include_field_for_diff_computation(field)

    def include_for_live_patch(self, context: LivePatchContext) -> bool:
        """
        Indicates if this ModelObject should be considered when generating a live patch.

        This method can be overridden if certain object should be ignored depending on their values,
        e.g. secrets that contain only a dummy secret
        """
        return True

    def include_existing_object_for_live_patch(self, org_id: str, parent_object: ModelObject | None) -> bool:
        """
        Indicates if this live ModelObject should be considered when generating a live patch.

        This method can be overridden if certain live objects should be ignored in some cases, e.g. environments
        """
        return True

    def keys(
        self,
        for_diff: bool = False,
        for_patch: bool = False,
        include_model_only_fields: bool = False,
        include_nested_models: bool = False,
        exclude_unset_keys: bool = True,
    ) -> list[str]:
        result = []

        for field in self.model_fields():
            if for_diff is True and not self.include_field_for_diff_computation(field):
                continue

            if for_patch is True and not self.include_field_for_patch_computation(field):
                continue

            if (for_diff or for_patch) and not include_model_only_fields and self.is_model_only(field):
                continue

            if include_nested_models is False and self.is_nested_model(field):
                continue

            if exclude_unset_keys:
                value = self.__getattribute__(field.name)
                if not is_unset(value):
                    result.append(field.name)
            else:
                result.append(field.name)

        return result

    def to_model_dict(
        self,
        for_diff: bool = False,
        include_model_only_fields: bool = False,
        include_nested_models: bool = False,
        exclude_none_values: bool = False,
    ) -> dict[str, Any]:
        result = {}

        for key in self.keys(
            for_diff=for_diff,
            include_model_only_fields=include_model_only_fields,
            include_nested_models=include_nested_models,
            exclude_unset_keys=True,
        ):
            value = self.__getattribute__(key)
            if exclude_none_values and not is_set_and_valid(value):
                continue
            elif self.is_nested_model_key(key):
                result[key] = cast(ModelObject, value).to_model_dict(for_diff, include_nested_models)
            elif self.is_embedded_model_key(key) and is_set_and_valid(value):
                result[key] = cast(EmbeddedModelObject, value).to_model_dict()
            else:
                result[key] = value

        return result

    def requires_web_ui(self) -> bool:
        return False

    def changes_require_web_ui(self, changes: dict[str, Change]) -> bool:
        return False

    def unset_settings_requiring_web_ui(self) -> None:
        return

    def contains_secrets(self) -> bool:
        return False

    def resolve_secrets(self, secret_resolver: Callable[[str], str]) -> None:
        return

    def copy_secrets(self, other_object: ModelObject) -> None:
        return

    def update_dummy_secrets(self, new_value: str) -> None:
        return

    @abstractmethod
    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> str | None: ...

    def to_jsonnet(
        self,
        printer: IndentingPrinter,
        jsonnet_config: JsonnetConfig,
        context: PatchContext,
        extend: bool,
        default_object: ModelObject,
    ) -> None:
        patch = self.get_patch_to(default_object)

        template_function = self.get_jsonnet_template_function(jsonnet_config, False)

        if self.is_keyed():
            key = patch.pop(self.get_key())
            printer.print(f"{unwrap(template_function)}('{key}')")
        else:
            printer.print(f"{unwrap(template_function)}")

        write_patch_object_as_json(patch, printer)

    @classmethod
    def generate_live_patch(
        cls: type[MT],
        expected_object: MT | None,
        current_object: MT | None,
        parent_object: ModelObject | None,
        context: LivePatchContext,
        handler: LivePatchHandler,
    ) -> None:
        if current_object is None:
            expected_object = unwrap(expected_object)
            handler(LivePatch.of_addition(expected_object, parent_object, expected_object.apply_live_patch))
            return

        if expected_object is None:
            current_object = unwrap(current_object)
            handler(LivePatch.of_deletion(current_object, parent_object, current_object.apply_live_patch))
            return

        modified_rule: dict[str, Change[Any]] = expected_object.get_difference_from(current_object)

        if len(modified_rule) > 0:
            handler(
                LivePatch.of_changes(
                    expected_object,
                    current_object,
                    modified_rule,
                    parent_object,
                    False,
                    expected_object.apply_live_patch,
                )
            )

    @classmethod
    def generate_live_patch_of_list(
        cls,
        expected_objects: Sequence[MT],
        current_objects: Sequence[MT],
        parent_object: MT | None,
        context: LivePatchContext,
        handler: LivePatchHandler,
    ) -> None:
        expected_objects_by_key = associate_by_key(expected_objects, lambda x: x.get_key_value())
        expected_objects_by_all_keys = multi_associate_by_key(expected_objects, lambda x: x.get_all_key_values())

        has_wildcard_keys = any(x.get_key_value().endswith("*") for x in expected_objects)

        for current_object in current_objects:
            key = current_object.get_key_value()

            expected_object = expected_objects_by_all_keys.get(key)
            if expected_object is None and has_wildcard_keys:
                for obj in expected_objects:
                    stripped_key = obj.get_key_value().rstrip("*")
                    if stripped_key and key.startswith(stripped_key):
                        expected_object = obj
                        break

            if expected_object is None:
                if current_object.include_existing_object_for_live_patch(context.org_id, parent_object):
                    cls.generate_live_patch(None, current_object, parent_object, context, handler)
                continue

            if expected_object.include_for_live_patch(context):
                cls.generate_live_patch(expected_object, current_object, parent_object, context, handler)

            for k in expected_object.get_all_key_values():
                expected_objects_by_all_keys.pop(k)
            expected_objects_by_key.pop(expected_object.get_key_value())

        for _, expected_object in expected_objects_by_key.items():
            if expected_object.include_for_live_patch(context):
                cls.generate_live_patch(expected_object, None, parent_object, context, handler)

    @classmethod
    @abstractmethod
    async def apply_live_patch(cls: type[MT], patch: LivePatch[MT], org_id: str, provider: GitHubProvider) -> None: ...



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from otterdog.providers.github import GitHubProvider
from otterdog.utils import get_approval

from . import Operation

if TYPE_CHECKING:
    from typing import Any

    from otterdog.config import OrganizationConfig


class ReviewAppPermissionsOperation(Operation):
    """
    Reviews permission updates of installed GitHub apps for an organization.
    """

    def __init__(self, app_slug: str | None, grant: bool, force: bool):
        super().__init__()
        self._app_slug = app_slug
        self._grant = grant
        self._force = force

    @property
    def app_slug(self) -> str | None:
        return self._app_slug

    @property
    def grant(self) -> bool:
        return self._grant

    @property
    def force(self) -> bool:
        return self._force

    def pre_execute(self) -> None:
        self.printer.println("Reviewing permission updates for app installations:")

    async def execute(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> int:
        github_id = org_config.github_id

        self._print_project_header(org_config, org_index, org_count)
        self.printer.level_up()

        try:
            try:
                credentials = self.get_credentials(org_config, only_token=False)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{e!s}")
                return 1

            async with GitHubProvider(credentials) as provider:
                apps = await provider.rest_api.org.get_app_installations(github_id)

                async with provider.web_client.get_logged_in_page() as page:
                    requested_permissions = await provider.web_client.get_requested_permission_updates(github_id, page)

                    for installation_id, permissions in requested_permissions.items():
                        app_slug = _get_app_slug_by_installation_id(apps, installation_id)
                        if app_slug is None:
                            self.printer.print_error(f"failed to process app installation with id '{installation_id}'")
                            continue

                        if self.app_slug is not None and self.app_slug != app_slug:
                            continue

                        self.printer.println()
                        self.print_dict(permissions, f"app['{app_slug}']", "", "black")

                        if self.grant is True:
                            if self.force is False:
                                self.printer.println()
                                self.printer.println("[bold]Approve[/] requested permissions?")
                                self.printer.println(
                                    "  Do you want to continue? (Only 'yes' or 'y' will be accepted to approve)\n"
                                )

                                self.printer.print("[bold]Enter a value:[/] ")
                                if not get_approval():
                                    self.printer.println("\nApproval cancelled.")
                                    continue

                            await provider.web_client.approve_requested_permission_updates(
                                github_id,
                                installation_id,
                                page,
                            )
                            self.printer.println()
                            self.printer.println("requested permissions approved.")

            return 0

        finally:
            self.printer.level_down()


def _get_app_slug_by_installation_id(installed_apps: list[dict[str, Any]], install_id: str) -> str | None:
    for app in installed_apps:
        if str(app["id"]) == install_id:
            return app["app_slug"]

    return None



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************


import yaml

from otterdog.webapp.blueprints import BlueprintType, read_blueprint
from otterdog.webapp.blueprints.append_configuration import AppendConfigurationBlueprint

yaml_content = """
id: prevent-force-pushes
name: Prevents force-pushes for the default branch
type: append_configuration
config:
  condition: >-
    $count($.rulesets[allows_force_pushes = false and
    "~DEFAULT_BRANCH" in include_refs and
    "~ALL" in include_repo_names and target = "branch"]) > 0
  content: |-
    {
      # snippet added due to '{{blueprint_url}}'
      rulesets+: [
        orgs.newOrgRuleset('{{blueprint_id}}') {
          allows_creations: true,
          include_repo_names: [
            "~ALL"
          ],
          include_refs: [
            "~DEFAULT_BRANCH"
          ],
          required_pull_request: null,
          required_status_checks: null,
        },
      ],
    }
"""


def test_read():
    config = yaml.safe_load(yaml_content)

    blueprint = read_blueprint("a", config)
    assert blueprint.type == BlueprintType.APPEND_CONFIGURATION

    assert isinstance(blueprint, AppendConfigurationBlueprint)
    assert "\n" not in blueprint.condition



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from abc import ABC, abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING

from otterdog.config import CredentialResolver
from otterdog.utils import Change, IndentingPrinter, unwrap

if TYPE_CHECKING:
    from typing import Any

    from otterdog.config import OrganizationConfig, OtterdogConfig
    from otterdog.credentials import Credentials


class Operation(ABC):
    _DEFAULT_WIDTH: int = 33

    def __init__(self) -> None:
        self._config: OtterdogConfig | None = None
        self._printer: IndentingPrinter | None = None

    def init(self, config: OtterdogConfig, printer: IndentingPrinter) -> None:
        self._config = config
        self._printer = printer

    @property
    def config(self) -> OtterdogConfig:
        return unwrap(self._config)

    @cached_property
    def credential_resolver(self) -> CredentialResolver:
        return CredentialResolver(self.config)

    def get_credentials(self, org_config: OrganizationConfig, only_token: bool = False) -> Credentials:
        return self.credential_resolver.get_credentials(org_config, only_token)

    @property
    def printer(self) -> IndentingPrinter:
        return unwrap(self._printer)

    @printer.setter
    def printer(self, value: IndentingPrinter):
        self._printer = value

    @abstractmethod
    def pre_execute(self) -> None: ...

    @abstractmethod
    async def execute(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> int: ...

    def post_execute(self) -> None:
        return

    async def check_config_file_exists(self, file_name: str) -> bool:
        from aiofiles import ospath

        if not await ospath.exists(file_name):
            self.printer.print_error(
                f"configuration file '{file_name}' does not exist, run 'fetch-config' or 'import' first."
            )
            return False
        else:
            return True

    async def check_config_file_overwrite_if_exists(self, file_name: str, force: bool) -> bool:
        from aiofiles import ospath

        from otterdog.utils import get_approval

        if await ospath.exists(file_name) and not force:
            self.printer.println()
            self.printer.println(f"[bold]Configuration already exists[/] at '{file_name}'.")
            self.printer.println("  Performing this action will overwrite its contents.")
            self.printer.println("  Do you want to continue? (Only 'yes' or 'y' will be accepted to approve)\n")

            self.printer.print("[bold]Enter a value:[/] ")
            if not get_approval():
                self.printer.println("\nOperation cancelled.")
                return False

        return True

    def _print_project_header(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> None:
        self.printer.println(
            f"\nProject [bold]{org_config.name}[/]\\[github_id={org_config.github_id}]"
            f"{self._format_progress(org_index, org_count)}"
        )

    @staticmethod
    def _format_progress(org_index: int | None, org_count: int | None) -> str:
        if org_index is None or org_count is None:
            return ""
        else:
            return f" ({org_index}/{org_count})"

    def print_dict(
        self,
        data: dict[str, Any],
        item_header: str,
        action: str,
        color: str,
        key_value_separator: str = "=",
        value_separator: str = "",
    ) -> None:
        prefix = f"[{color}]{action}[/] " if action else ""
        closing_prefix = prefix

        max_key_length = self._get_max_key_length_for_nested_dict(data)

        if item_header:
            self.printer.print(f"{prefix}{item_header} ")
        self._print_dict_internal(
            data, max_key_length, prefix, closing_prefix, False, key_value_separator, value_separator
        )

    def _print_internal(
        self,
        data: Any,
        max_key_length: int,
        prefix: str,
        closing_prefix: str,
        include_prefix: bool,
        key_value_separator: str,
        value_separator: str,
    ):
        if isinstance(data, dict):
            self._print_dict_internal(
                data,
                max_key_length,
                prefix,
                closing_prefix,
                include_prefix,
                key_value_separator,
                value_separator,
            )
        elif isinstance(data, list):
            self._print_list_internal(
                data,
                max_key_length,
                prefix,
                closing_prefix,
                key_value_separator,
                value_separator,
            )
        else:
            self.printer.println(
                f"{prefix if include_prefix else ''}{self._get_value(data)}{value_separator}", highlight=True
            )

    def _print_dict_internal(
        self,
        data: dict[str, Any],
        max_key_length: int,
        prefix: str,
        closing_prefix: str,
        include_prefix: bool,
        key_value_separator: str,
        value_separator: str,
    ):
        if len(data) > 0:
            self.printer.println(f"{prefix if include_prefix else ''}{{")
            self.printer.level_up()
            for key, value in sorted(data.items()):
                self.printer.print(f"{prefix}{key.ljust(max_key_length, ' ')} {key_value_separator} ")
                self._print_internal(
                    value, max_key_length, prefix, closing_prefix, False, key_value_separator, value_separator
                )

            self.printer.level_down()
            self.printer.println(f"{closing_prefix}}}{value_separator}")
        else:
            self.printer.println(f"{prefix if include_prefix else ''}{{}}{value_separator}")

    def _print_list_internal(
        self,
        data: list[Any],
        max_key_length: int,
        prefix: str,
        closing_prefix: str,
        key_value_separator: str,
        value_separator: str,
    ):
        if len(data) > 0:
            self.printer.println("[", highlight=True)
            self.printer.level_up()
            for entry in data:
                self._print_internal(
                    entry,
                    max_key_length,
                    prefix,
                    closing_prefix,
                    True,
                    key_value_separator,
                    value_separator,
                )

            self.printer.level_down()
            self.printer.println(f"{closing_prefix}],", highlight=True)
        else:
            self.printer.println(f"[]{value_separator}")

    @staticmethod
    def _get_value(value: Any) -> str:
        if value is None:
            return "null"
        if isinstance(value, bool):
            return str(value).lower()
        elif isinstance(value, int):
            return str(value)
        else:
            return f'"{value}"'

    def _get_max_key_length(self, data: dict[str, Change[Any]]) -> int:
        length = 0

        for key, change in sorted(data.items()):
            expected_value = change.to_value

            length = max(length, len(key))
            if isinstance(expected_value, dict):
                for k, _ in expected_value.items():
                    length = max(length, len(k) + self.printer.spaces_per_level)

        return length

    def _get_max_key_length_for_nested_dict(self, data: dict[str, Any]) -> int:
        length = 0

        for key, value in sorted(data.items()):
            length = max(length, len(key))
            if isinstance(value, dict):
                for k, _ in value.items():
                    length = max(length, len(k) + self.printer.spaces_per_level)

        return length

    def print_modified_dict(
        self,
        data: dict[str, Change[Any]],
        item_header: str,
        forced_update: bool = False,
    ) -> None:
        prefix = "[magenta]! [/]" if forced_update else "[yellow]~ [/]"
        closing_prefix = prefix
        color = "magenta" if forced_update else "yellow"

        self.printer.println(f"\n{prefix}{item_header} {{")
        self.printer.level_up()

        max_key_length = self._get_max_key_length(data)

        for key, change in sorted(data.items()):
            current_value = change.from_value
            expected_value = change.to_value
            self._print_modified_internal(
                key,
                max_key_length,
                current_value,
                expected_value,
                prefix,
                color,
                forced_update,
            )

        self.printer.level_down()
        self.printer.println(f"{closing_prefix}}}")

    def _print_modified_internal(
        self,
        key: str,
        max_key_length: int,
        current_value,
        expected_value,
        prefix: str,
        color: str,
        forced_update: bool,
    ):
        if isinstance(expected_value, dict):
            self.printer.println(f"{prefix}{key.ljust(max_key_length, ' ')} = {{")
            self.printer.level_up()
            self._print_modified_dict_internal(
                max_key_length,
                current_value,
                expected_value,
                prefix,
                color,
                forced_update,
            )
            self.printer.level_down()
            self.printer.println(f"{prefix}}}")
        elif isinstance(expected_value, list):
            self.printer.println(f"{prefix}{key.ljust(max_key_length, ' ')} = [")
            self.printer.level_up()
            self._print_modified_list_internal(current_value, expected_value, prefix, color, forced_update)
            self.printer.level_down()
            self.printer.println(f"{prefix}]")
        else:
            self.printer.println(
                f"{prefix}{key.ljust(max_key_length, ' ')} ="
                f' {self._get_value(current_value)} [{color}]->[/] {self._get_value(expected_value)}'
            )

    def _print_modified_dict_internal(
        self,
        max_key_length: int,
        current_value,
        expected_value,
        prefix: str,
        color: str,
        forced_update: bool,
    ) -> None:
        processed_keys = set()
        for k, v in sorted(expected_value.items()):
            c_v = current_value.get(k) if current_value is not None else None

            if v != c_v:
                if c_v is None:
                    self.printer.println(f"[green]+ [/]{k.ljust(max_key_length, ' ')} =" f" {self._get_value(v)}")
                elif v is None:
                    self.printer.println(f"[red]- [/]{k.ljust(max_key_length, ' ')} =" f" {self._get_value(c_v)}")
                else:
                    self._print_modified_internal(k, max_key_length, c_v, v, prefix, color, forced_update)
            elif forced_update is True:
                self._print_modified_internal(k, max_key_length, c_v, v, prefix, color, forced_update)

            processed_keys.add(k)

        if current_value is not None:
            for k, v in sorted(current_value.items()):
                if k not in processed_keys:
                    self.printer.println(f"[red]- [/]{k.ljust(max_key_length, ' ')} =" f" {self._get_value(v)}")

    def _print_modified_list_internal(
        self,
        current_value,
        expected_value,
        prefix: str,
        color: str,
        forced_update: bool,
    ) -> None:
        from difflib import SequenceMatcher

        a = sorted(current_value)
        b = sorted(expected_value)

        max_length_a = max((len(str(x)) for x in a), default=0)
        max_length_b = max((len(str(x)) for x in b), default=0)

        max_length = max(max_length_a, max_length_b)

        for tag, i1, i2, j1, j2 in SequenceMatcher(None, a, b).get_opcodes():
            match tag:
                case "equal":
                    if forced_update is True:
                        for i in range(i1, i2):
                            self.printer.println(f"{prefix}{self._get_value(a[i])}")
                case "replace":
                    diff_i = i2 - i1
                    diff_j = j2 - j1

                    i = i1
                    j = j1

                    for _ in range(i1, min(diff_i, diff_j)):
                        self.printer.println(
                            f"{prefix}{self._get_value(a[i]).ljust(max_length, ' ')} [{color}]->[/] "
                            f"{self._get_value(b[j])}"
                        )
                        j = j + 1
                        i = i + 1

                    while i < i2:
                        self.printer.println(f"[red]- [/]{self._get_value(a[i])}")
                        i = i + 1

                    while j < j2:
                        self.printer.println(f"[green]+ [/]{self._get_value(b[j])}")
                        j = j + 1

                case "delete":
                    for i in range(i1, i2):
                        self.printer.println(f"[red]- [/]{self._get_value(a[i])}")
                case "insert":
                    for j in range(j1, j2):
                        self.printer.println(f"[green]+ [/]{self._get_value(b[j])}")



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from collections.abc import Mapping
from typing import Any

from otterdog.jsonnet import JsonnetConfig
from otterdog.models import ModelObject
from otterdog.models.environment import Environment
from otterdog.utils import UNSET, query_json

from . import ModelTest


class EnvironmentTest(ModelTest):
    def create_model(self, data: Mapping[str, Any]) -> ModelObject:
        return Environment.from_model_data(data)

    @property
    def template_function(self) -> str:
        return JsonnetConfig.create_environment

    @property
    def model_data(self):
        return self.load_json_resource("otterdog-environment.json")

    @property
    def provider_data(self):
        return self.load_json_resource("github-environment.json")

    def test_load_from_model(self):
        env = Environment.from_model_data(self.model_data)

        assert env.id is UNSET
        assert env.node_id is UNSET
        assert env.name == "linux"
        assert env.wait_timer == 15
        assert env.reviewers == ["@netomi", "@OtterdogTest/eclipsefdn-security"]
        assert env.deployment_branch_policy == "selected"
        assert env.branch_policies == ["main", "develop/*"]

    def test_load_from_provider(self):
        env = Environment.from_provider_data(self.org_id, self.provider_data)

        assert env.id == 1102681190
        assert env.node_id == "EN_kwDOI9xAhM5BuZRm"
        assert env.name == "linux"
        assert env.wait_timer == 15
        assert env.reviewers == ["@netomi", "@OtterdogTest/eclipsefdn-security"]
        assert env.deployment_branch_policy == "selected"
        assert env.branch_policies == ["main", "develop/*"]

    async def test_to_provider(self):
        env = Environment.from_model_data(self.model_data)

        provider_data = await env.to_provider_data(self.org_id, self.provider)

        assert len(provider_data) == 5
        assert provider_data["wait_timer"] == 15

        assert query_json("reviewers[0].id", provider_data) == "id_netomi"
        assert query_json("reviewers[1].id", provider_data) == "id_OtterdogTest/eclipsefdn-security"

        assert query_json("deployment_branch_policy.protected_branches", provider_data) is False
        assert query_json("deployment_branch_policy.custom_branch_policies", provider_data) is True



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import yaml

from otterdog.webapp.policies import PolicyType, create_policy, read_policy
from otterdog.webapp.policies.macos_large_runners import MacOSLargeRunnersUsagePolicy

disabled = """
name: Limit use of macos larger runners
type: macos_large_runners
config:
  allowed: false
"""

enabled = """
name: Limit use of macos larger runners
type: macos_large_runners
config:
  allowed: true
"""


def test_read_disabled():
    config = yaml.safe_load(disabled)

    policy = read_policy("a", config)
    assert policy.type == PolicyType.MACOS_LARGE_RUNNERS_USAGE

    assert isinstance(policy, MacOSLargeRunnersUsagePolicy)
    assert policy.path == "a"
    assert policy.name == "Limit use of macos larger runners"
    assert policy.description is None
    assert policy._is_workflow_job_permitted([]) == (False, True)
    assert policy._is_workflow_job_permitted(["macos-runners-large"]) == (True, False)


def test_read_enabled():
    config = yaml.safe_load(enabled)

    policy = read_policy("a", config)
    assert policy.type == PolicyType.MACOS_LARGE_RUNNERS_USAGE

    assert isinstance(policy, MacOSLargeRunnersUsagePolicy)
    assert policy.path == "a"
    assert policy.name == "Limit use of macos larger runners"
    assert policy.description is None
    assert policy._is_workflow_job_permitted([]) == (False, True)
    assert policy._is_workflow_job_permitted(["macos-runners-large"]) == (True, True)


def test_create():
    config = yaml.safe_load(enabled)

    policy = create_policy(PolicyType.MACOS_LARGE_RUNNERS_USAGE, "a", "myname", "mydesc", config["config"])

    assert policy.type == PolicyType.MACOS_LARGE_RUNNERS_USAGE

    assert isinstance(policy, MacOSLargeRunnersUsagePolicy)
    assert policy.path == "a"
    assert policy.name == "myname"
    assert policy.description == "mydesc"
    assert policy._is_workflow_job_permitted([]) == (False, True)
    assert policy._is_workflow_job_permitted(["macos-runners-large"]) == (True, True)


def test_merge():
    disabled_config = yaml.safe_load(disabled)
    enabled_config = yaml.safe_load(enabled)

    global_policy = read_policy("global", disabled_config)
    local_policy = read_policy("local", enabled_config)

    merged_policy = global_policy.merge(local_policy)

    assert merged_policy.type == PolicyType.MACOS_LARGE_RUNNERS_USAGE

    assert isinstance(merged_policy, MacOSLargeRunnersUsagePolicy)
    assert merged_policy.path == "local"
    assert merged_policy.name == "Limit use of macos larger runners"
    assert merged_policy.description is None
    assert merged_policy._is_workflow_job_permitted([]) == (False, True)
    assert merged_policy._is_workflow_job_permitted(["macos-runners-large"]) == (True, True)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from aiofiles import open, os, ospath

from otterdog.providers.github import GitHubProvider

from . import Operation

if TYPE_CHECKING:
    from otterdog.config import OrganizationConfig


class FetchOperation(Operation):
    """
    Fetches the current configuration from the meta-data repository of an organization.
    """

    def __init__(self, force_processing: bool, pull_request: str, suffix: str = "", ref: str | None = None):
        super().__init__()
        self._force_processing = force_processing
        self._pull_request = pull_request
        self._suffix = suffix
        self._ref = ref

    @property
    def force_processing(self) -> bool:
        return self._force_processing

    @property
    def pull_request(self) -> str:
        return self._pull_request

    @property
    def suffix(self) -> str:
        return self._suffix

    @property
    def ref(self) -> str | None:
        return self._ref

    def pre_execute(self) -> None:
        self.printer.println("Fetching organization configurations:")

    async def execute(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config

        self._print_project_header(org_config, org_index, org_count)

        org_file_name = jsonnet_config.org_config_file + self.suffix
        if not await self.check_config_file_overwrite_if_exists(org_file_name, self.force_processing):
            return 1

        self.printer.level_up()

        try:
            try:
                credentials = self.get_credentials(org_config, only_token=True)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{e!s}")
                return 1

            async with GitHubProvider(credentials) as provider:
                try:
                    if self.pull_request is not None:
                        ref = await provider.get_ref_for_pull_request(
                            org_config.github_id, org_config.config_repo, self.pull_request
                        )
                    else:
                        ref = self.ref

                    definition = await provider.get_content(
                        org_config.github_id,
                        org_config.config_repo,
                        f"otterdog/{github_id}.jsonnet",
                        ref,
                    )
                except RuntimeError:
                    self.printer.print_error(f"failed to fetch definition from repo '{org_config.config_repo}'")
                    return 1

            output_dir = jsonnet_config.org_dir
            if not await ospath.exists(output_dir):
                await os.makedirs(output_dir)

            async with open(org_file_name, "w") as file:
                await file.write(definition)

            if self.pull_request is not None:
                self.printer.println(
                    f"organization definition fetched from pull request " f"#{self.pull_request} to '{org_file_name}'"
                )
            else:
                self.printer.println(f"organization definition fetched from default branch to '{org_file_name}'")

            return 0
        finally:
            self.printer.level_down()



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from otterdog.logging import is_info_enabled
from otterdog.models import FailureType
from otterdog.models.github_organization import GitHubOrganization
from otterdog.providers.github import GitHubProvider

from . import Operation

if TYPE_CHECKING:
    from otterdog.config import OrganizationConfig
    from otterdog.jsonnet import JsonnetConfig


class ValidateOperation(Operation):
    """
    Validates local organization configurations.
    """

    def __init__(self):
        super().__init__()

    def pre_execute(self) -> None:
        self.printer.println("Validating organization configurations:")

    async def execute(
        self,
        org_config: OrganizationConfig,
        org_index: int | None = None,
        org_count: int | None = None,
    ) -> int:
        github_id = org_config.github_id
        jsonnet_config = org_config.jsonnet_config
        await jsonnet_config.init_template()

        self._print_project_header(org_config, org_index, org_count)
        self.printer.level_up()

        try:
            org_file_name = jsonnet_config.org_config_file
            if not await self.check_config_file_exists(org_file_name):
                return 1

            try:
                organization = GitHubOrganization.load_from_file(github_id, org_file_name)
            except RuntimeError as ex:
                self.printer.print_error(f"Validation failed\nfailed to load configuration: {ex!s}")
                return 1

            try:
                credentials = self.get_credentials(org_config, only_token=True)
            except RuntimeError as e:
                self.printer.print_error(f"invalid credentials\n{e!s}")
                return 1

            async with GitHubProvider(credentials) as provider:
                validation_infos, validation_warnings, validation_errors = await self.validate(
                    organization, jsonnet_config, provider
                )
                validation_count = validation_infos + validation_warnings + validation_errors

            if validation_count == 0:
                self.printer.println("[green]Validation succeeded[/]")
            else:
                if validation_errors == 0:
                    self.printer.println(
                        f"[green]Validation succeeded[/]': "
                        f"{validation_infos} info(s), {validation_warnings} warning(s), "
                        f"{validation_errors} error(s)"
                    )
                else:
                    self.printer.println(
                        f"[red]Validation failed[/]: "
                        f"{validation_infos} info(s), {validation_warnings} warning(s), "
                        f"{validation_errors} error(s)"
                    )

            if validation_infos > 0 and not is_info_enabled():
                self.printer.level_up()
                self.printer.println(
                    "in order to print validation infos, enable printing info messages by adding '-v' flag."
                )
                self.printer.level_down()

            return validation_errors
        finally:
            self.printer.level_down()

    async def validate(
        self,
        organization: GitHubOrganization,
        jsonnet_config: JsonnetConfig,
        provider: GitHubProvider,
    ) -> tuple[int, int, int]:
        if organization.secrets_resolved is True:
            raise RuntimeError("validation requires an unresolved model.")

        context = await organization.validate(self.config, jsonnet_config, self.credential_resolver, provider)

        validation_infos = 0
        validation_warnings = 0
        validation_errors = 0

        for failure_type, message in context.validation_failures:
            match failure_type:
                case FailureType.INFO:
                    self.printer.print_info(message)
                    validation_infos += 1

                case FailureType.WARNING:
                    self.printer.print_warn(message)
                    validation_warnings += 1

                case FailureType.ERROR:
                    self.printer.print_error(message)
                    validation_errors += 1

        return validation_infos, validation_warnings, validation_errors



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from aiofiles import ospath

from otterdog.models.github_organization import GitHubOrganization
from otterdog.providers.github import GitHubProvider
from otterdog.utils import unwrap

from .plan import PlanOperation

if TYPE_CHECKING:
    from otterdog.config import OrganizationConfig
    from otterdog.jsonnet import JsonnetConfig


class LocalPlanOperation(PlanOperation):
    def __init__(
        self,
        suffix: str,
        repo_filter: str,
        update_webhooks: bool,
        update_secrets: bool,
        update_filter: str,
    ) -> None:
        super().__init__(
            no_web_ui=False,
            repo_filter=repo_filter,
            update_webhooks=update_webhooks,
            update_secrets=update_secrets,
            update_filter=update_filter,
        )

        self.suffix = suffix
        self._other_org: GitHubOrganization | None = None

    @property
    def other_org(self) -> GitHubOrganization:
        return unwrap(self._other_org)

    def pre_execute(self) -> None:
        self.printer.println("Printing local diff:")
        self.print_legend()

    def verbose_output(self):
        return False

    def resolve_secrets(self) -> bool:
        return False

    def coerce_current_org(self) -> bool:
        return True

    def setup_github_client(self, org_config: OrganizationConfig) -> GitHubProvider:
        return GitHubProvider(self.get_credentials(org_config, only_token=True))

    async def load_current_org(
        self, project_name: str, github_id: str, jsonnet_config: JsonnetConfig
    ) -> GitHubOrganization:
        other_org_file_name = jsonnet_config.org_config_file + self.suffix

        if not await ospath.exists(other_org_file_name):
            raise RuntimeError(f"configuration file '{other_org_file_name}' does not exist")

        return GitHubOrganization.load_from_file(github_id, other_org_file_name)

    def preprocess_orgs(
        self, expected_org: GitHubOrganization, current_org: GitHubOrganization
    ) -> tuple[GitHubOrganization, GitHubOrganization]:
        expected_org.update_dummy_secrets("<DUMMY>")
        current_org.update_dummy_secrets("<DUMMY>")
        return expected_org, current_org



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import json
from collections.abc import AsyncIterable, Mapping
from typing import Any

from aiohttp import ClientSession, ClientTimeout, TCPConnector
from aiohttp_client_cache.session import CachedSession as AsyncCachedSession
from aiohttp_retry import ExponentialRetry, RetryClient

from otterdog.logging import is_trace_enabled
from otterdog.providers.github.auth import AuthStrategy
from otterdog.providers.github.cache import CacheStrategy
from otterdog.providers.github.exception import (
    BadCredentialsException,
    GitHubException,
    InsufficientPermissionsException,
)
from otterdog.providers.github.stats import RequestStatistics
from otterdog.utils import get_logger

_logger = get_logger(__name__)


class Requester:
    def __init__(
        self,
        auth_strategy: AuthStrategy | None,
        cache_strategy: CacheStrategy,
        base_url: str,
        api_version: str,
    ):
        self._auth = auth_strategy.get_auth() if auth_strategy is not None else None

        self._headers = {
            "Accept": "application/vnd.github+json",
            "X-GitHub-Api-Version": api_version,
            "X-Github-Next-Global-ID": "1",
        }

        self._statistics = RequestStatistics()
        self._cache_strategy = cache_strategy

        if self._cache_strategy.is_external():
            self._base_url = cache_strategy.replace_base_url(f"https://{base_url}")
            self._session = ClientSession()
        else:
            self._base_url = f"https://{base_url}"
            self._session = AsyncCachedSession(
                cache=self._cache_strategy.get_cache_backend(),
                timeout=ClientTimeout(connect=3, sock_connect=3),
                connector=TCPConnector(
                    limit=30,
                ),
            )

        self._client = RetryClient(
            retry_options=ExponentialRetry(3, exceptions={Exception}),
            client_session=self._session,
        )

    @property
    def statistics(self) -> RequestStatistics:
        return self._statistics

    async def close(self) -> None:
        await self._session.close()

    def _build_url(self, url_path: str) -> str:
        return f"{self._base_url}{url_path}"

    async def request_paged_json(
        self,
        method: str,
        url_path: str,
        data: dict[str, Any] | None = None,
        params: dict[str, str] | None = None,
        entries_key: str | None = None,
    ) -> list[dict[str, Any]]:
        from urllib import parse

        json_data = None
        if data is not None:
            json_data = json.dumps(data)

        result = []
        query_params: dict[str, str] | None = {"per_page": "100"}

        while query_params is not None:
            if params is not None:
                query_params.update(params)

            status, body, next_url, _ = await self._request_raw_with_next_link(
                method, url_path, json_data, query_params
            )
            self._check_response(url_path, status, body)
            response = json.loads(body)

            entries = response[entries_key] if entries_key is not None else response

            if next_url is None:
                query_params = None
            else:
                query_params = {k: v[0] for k, v in parse.parse_qs(parse.urlparse(next_url).query).items()}

            for item in entries:
                result.append(item)

        return result

    async def request_json(
        self,
        method: str,
        url_path: str,
        data: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
    ) -> Any:
        input_data = None
        if data is not None:
            input_data = json.dumps(data)

        status, body = await self.request_raw(method, url_path, input_data, params)
        self._check_response(url_path, status, body)
        json_result = json.loads(body)
        if is_trace_enabled():
            _logger.trace("'%s' url = %s, json = %s", method, url_path, json.dumps(json_result, indent=2))
        return json_result

    async def request_raw(
        self,
        method: str,
        url_path: str,
        data: str | None = None,
        params: dict[str, Any] | None = None,
    ) -> tuple[int, str]:
        status, body, _, _ = await self._request_raw_with_next_link(method, url_path, data, params)
        _logger.trace("'%s' url = %s, result = (%d)", method, url_path, status)
        return status, body

    async def request_raw_with_scopes(
        self,
        method: str,
        url_path: str,
        data: str | None = None,
        params: dict[str, Any] | None = None,
    ) -> tuple[int, str, str]:
        status, body, _, scopes = await self._request_raw_with_next_link(method, url_path, data, params)
        _logger.trace("'%s' url = %s, result = (%d)", method, url_path, status)
        return status, body, scopes

    async def _request_raw_with_next_link(
        self,
        method: str,
        url_path: str,
        data: str | None = None,
        params: dict[str, Any] | None = None,
    ) -> tuple[int, str, str | None, str]:
        _logger.trace("'%s' url = %s, data = %s, params = %s", method, url_path, data, params)

        headers = self._headers.copy()
        if self._auth is not None:
            self._auth.update_headers_with_authorization(headers)

        url = self._build_url(url_path)
        async with self._client.request(
            method,
            url=url,
            headers=headers,
            params=params,
            data=data,
            **self._cache_strategy.get_request_parameters(),
        ) as response:
            self._statistics.sent_request()

            text = await response.text()
            status = response.status
            links = response.links
            next_link = links.get("next", None) if links is not None else None
            next_url = next_link.get("url", None) if next_link is not None else None

            self._check_permissions(url_path, status, text, response.headers)

            if (hasattr(response, "from_cache") and response.from_cache) or response.headers.get(
                "X-From-Cache", 0
            ) == "1":
                self._statistics.received_cached_response()
            else:
                self._statistics.update_remaining_rate_limit(int(response.headers.get("x-ratelimit-remaining", -1)))

            return (
                status,
                text,
                str(next_url) if next_url is not None else None,
                response.headers.get("X-OAuth-Scopes", ""),
            )

    async def request_stream(
        self,
        method: str,
        url_path: str,
        data: str | None = None,
        params: dict[str, Any] | None = None,
    ) -> AsyncIterable[bytes]:
        _logger.trace(
            "stream '%s' url = %s, data = %s, headers = %s",
            method,
            url_path,
            data,
            self._headers,
        )

        headers = self._headers.copy()
        if self._auth is not None:
            self._auth.update_headers_with_authorization(headers)

        url = self._build_url(url_path)
        async with self._client.request(
            method,
            url=url,
            headers=headers,
            params=params,
            data=data,
            **self._cache_strategy.get_request_parameters(),
        ) as response:
            async for chunk, _ in response.content.iter_chunks():
                yield chunk

    def _check_response(self, url_path: str, status_code: int, body: str) -> None:
        if status_code >= 400:
            self._create_exception(self._build_url(url_path), status_code, body)

    @staticmethod
    def _check_permissions(url_path: str, status_code: int, body: str, headers: Mapping[str, str]):
        if status_code == 403:
            existing_scopes = {x.strip() for x in headers.get("X-OAuth-Scopes", "").split(",") if len(x) > 0}
            required_scopes = {x.strip() for x in headers.get("X-Accepted-OAuth-Scopes", "").split(",") if len(x) > 0}

            missing_scopes = required_scopes - existing_scopes
            if len(missing_scopes) > 0:
                raise InsufficientPermissionsException(url_path, status_code, body, list(missing_scopes))

    @staticmethod
    def _create_exception(url: str, status_code: int, body: str):
        if status_code == 401:
            raise BadCredentialsException(url, body)
        else:
            raise GitHubException(url, status_code, body)



#  *******************************************************************************
#  Copyright (c) 2023-2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Any

from jsonbender import OptionalS, S  # type: ignore

from otterdog.models import (
    FailureType,
    LivePatch,
    LivePatchContext,
    LivePatchHandler,
    LivePatchType,
    ModelObject,
    ValidationContext,
)
from otterdog.utils import Change, is_set_and_present, is_set_and_valid, unwrap

if TYPE_CHECKING:
    from otterdog.jsonnet import JsonnetConfig
    from otterdog.providers.github import GitHubProvider


@dataclasses.dataclass
class CustomProperty(ModelObject):
    """
    Represents a Custom Property defined in an Organization.
    """

    name: str = dataclasses.field(metadata={"key": True})
    value_type: str
    required: bool
    default_value: str | list[str] | None
    description: str | None
    allowed_values: list[str] | None

    @property
    def model_object_name(self) -> str:
        return "custom_property"

    def validate(self, context: ValidationContext, parent_object: Any) -> None:
        if is_set_and_valid(self.value_type):
            if self.value_type not in {"string", "single_select", "multi_select", "true_false"}:
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'value_type' set to '{self.value_type}', "
                    f"while only values ('string' | 'single_select' | 'multi_select' | 'true_false') are allowed.",
                )

            if (
                self.value_type in {"single_select", "multi_select"}
                and is_set_and_present(self.allowed_values)
                and len(self.allowed_values) == 0
            ):
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'value_type' set to '{self.value_type}', "
                    f"but no 'allowed_values' are defined.",
                )

            if (
                self.value_type in {"single_select", "multi_select"}
                and is_set_and_present(self.allowed_values)
                and len(self.allowed_values) > 200
            ):
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'value_type' set to '{self.value_type}', "
                    f"but more than 200 elements as 'allowed_values' are defined.",
                )

        if is_set_and_valid(self.required):
            if self.required is True:
                if not is_set_and_present(self.default_value):
                    context.add_failure(
                        FailureType.ERROR,
                        f"{self.get_model_header(parent_object)} has 'required' set to 'true', "
                        f"but no property 'default_value' is specified.",
                    )
            elif is_set_and_present(self.default_value):
                if isinstance(self.default_value, str) and len(self.default_value) > 0:
                    context.add_failure(
                        FailureType.ERROR,
                        f"{self.get_model_header(parent_object)} has 'required' set to 'false' "
                        f"but property 'default_value' is set to a non-empty value.",
                    )
                elif isinstance(self.default_value, list) and len(self.default_value) > 0:
                    context.add_failure(
                        FailureType.ERROR,
                        f"{self.get_model_header(parent_object)} has 'required' set to 'false' "
                        f"but property 'default_value' is set to a non-empty list.",
                    )

        if is_set_and_valid(self.default_value) and is_set_and_valid(self.value_type):
            if self.value_type == "single_select" and not isinstance(self.default_value, str):
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'value_type' set to '{self.value_type}', "
                    f"but 'default_value' contains a list of values '{self.default_value}'.",
                )

        if (
            is_set_and_present(self.default_value)
            and is_set_and_present(self.allowed_values)
            and len(self.allowed_values) > 0
        ):
            if isinstance(self.default_value, str) and self.default_value not in self.allowed_values:
                context.add_failure(
                    FailureType.ERROR,
                    f"{self.get_model_header(parent_object)} has 'default_value' set to '{self.default_value}', "
                    f"but it is not in the list of allowed values '{self.allowed_values}'.",
                )
            elif isinstance(self.default_value, list):
                for value in self.default_value:
                    if value not in self.allowed_values:
                        context.add_failure(
                            FailureType.ERROR,
                            f"{self.get_model_header(parent_object)} has a 'default_value' set to "
                            f"'{self.default_value}', "
                            f"but some of its elements are not in the list of allowed values '{self.allowed_values}'.",
                        )

    def include_field_for_diff_computation(self, field: dataclasses.Field) -> bool:
        if self.required is not True and field.name in ["default_value"]:
            return False

        if self.value_type not in {"single_select", "multi_select"} and field.name in ["allowed_values"]:
            return False

        return True

    def include_field_for_patch_computation(self, field: dataclasses.Field) -> bool:
        return True

    @classmethod
    def get_mapping_from_provider(cls, org_id: str, data: dict[str, Any]) -> dict[str, Any]:
        mapping = super().get_mapping_from_provider(org_id, data)
        mapping.update(
            {
                "name": S("property_name"),
                "allowed_values": OptionalS("allowed_values", default=[]),
            }
        )
        return mapping

    @classmethod
    async def get_mapping_to_provider(
        cls, org_id: str, data: dict[str, Any], provider: GitHubProvider
    ) -> dict[str, Any]:
        mapping = await super().get_mapping_to_provider(org_id, data, provider)

        if "name" in data:
            mapping.pop("name")

        return mapping

    def get_jsonnet_template_function(self, jsonnet_config: JsonnetConfig, extend: bool) -> str | None:
        return f"orgs.{jsonnet_config.create_org_custom_property}"

    @classmethod
    def generate_live_patch(
        cls,
        expected_object: CustomProperty | None,
        current_object: CustomProperty | None,
        parent_object: ModelObject | None,
        context: LivePatchContext,
        handler: LivePatchHandler,
    ) -> None:
        if current_object is None:
            expected_object = unwrap(expected_object)
            handler(LivePatch.of_addition(expected_object, parent_object, expected_object.apply_live_patch))
            return

        if expected_object is None:
            handler(LivePatch.of_deletion(current_object, parent_object, current_object.apply_live_patch))
            return

        modified_property: dict[str, Change[Any]] = expected_object.get_difference_from(current_object)

        if len(modified_property) > 0:
            if "value_type" in modified_property:
                raise RuntimeError(
                    f"trying to change 'value_type' to '{expected_object.value_type}' for "
                    f"{expected_object.get_model_header(parent_object)} which is not supported."
                )

            handler(
                LivePatch.of_changes(
                    expected_object,
                    current_object,
                    modified_property,
                    parent_object,
                    False,
                    expected_object.apply_live_patch,
                )
            )

    @classmethod
    async def apply_live_patch(cls, patch: LivePatch[CustomProperty], org_id: str, provider: GitHubProvider) -> None:
        match patch.patch_type:
            case LivePatchType.ADD:
                expected_object = unwrap(patch.expected_object)
                await provider.add_org_custom_property(
                    org_id,
                    expected_object.name,
                    await expected_object.to_provider_data(org_id, provider),
                )

            case LivePatchType.REMOVE:
                current_object = unwrap(patch.current_object)
                await provider.delete_org_custom_property(org_id, current_object.name)

            case LivePatchType.CHANGE:
                await provider.update_org_custom_property(
                    org_id,
                    unwrap(patch.current_object).name,
                    await unwrap(patch.expected_object).to_provider_data(org_id, provider),
                )



#  *******************************************************************************
#  Copyright (c) 2024 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import json
from typing import Any

from otterdog.logging import get_logger
from otterdog.providers.github.exception import GitHubException

from . import RestApi, RestClient

_logger = get_logger(__name__)


class CommitClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def get_commit(self, org_id: str, repo_name: str, ref: str) -> dict[str, Any]:
        _logger.debug("getting commit for ref '%s' from repo '%s/%s'", ref, org_id, repo_name)

        try:
            return await self.requester.request_json("GET", f"/repos/{org_id}/{repo_name}/commits/{ref}")
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving commit:\n{ex}") from ex

    async def get_commit_statuses(self, org_id: str, repo_name: str, ref: str) -> list[dict[str, Any]]:
        _logger.debug("getting commit statuses for ref '%s' from repo '%s/%s'", ref, org_id, repo_name)

        try:
            return await self.requester.request_paged_json("GET", f"/repos/{org_id}/{repo_name}/commits/{ref}/statuses")
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving commit statuses:\n{ex}") from ex

    async def create_commit_status(
        self,
        org_id: str,
        repo_name: str,
        sha: str,
        state: str,
        context: str,
        description: str | None = None,
        target_url: str | None = None,
    ) -> None:
        _logger.debug("creating a commit status for sha '%s' in repo '%s/%s'", sha, org_id, repo_name)

        data = {"state": state, "target_url": target_url, "description": description, "context": context}
        status, body = await self.requester.request_raw(
            "POST", f"/repos/{org_id}/{repo_name}/statuses/{sha}", data=json.dumps(data)
        )

        if status != 201:
            raise RuntimeError(f"failed creating commit status for '{org_id}/{repo_name}/{sha}'\n{status}: {body}")

        _logger.debug("created commit status for sha '%s' in repo '%s/%s'", sha, org_id, repo_name)



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

import json
from typing import Any

from otterdog.providers.github.exception import GitHubException
from otterdog.utils import get_logger

from . import RestApi, RestClient, encrypt_value

_logger = get_logger(__name__)


class OrgClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def get_id(self, org_id: str) -> int:
        _logger.debug("retrieving id for org '%s'", org_id)

        try:
            settings = await self.requester.request_json("GET", f"/orgs/{org_id}")
            return settings["id"]
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving id for org '{org_id}':\n{ex}") from ex

    async def get_settings(self, org_id: str, included_keys: set[str]) -> dict[str, Any]:
        _logger.debug("retrieving settings for org '%s'", org_id)

        try:
            settings = await self.requester.request_json("GET", f"/orgs/{org_id}")
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving settings for org '{org_id}':\n{ex}") from ex

        if "security_managers" in included_keys:
            security_manager_role_id = await self.get_role_id(org_id, "security_manager")
            # handle gracefully if the role does not exist
            if security_manager_role_id is None:
                settings["security_managers"] = []
            else:
                security_managers = await self.list_security_managers(org_id, str(security_manager_role_id))
                settings["security_managers"] = security_managers

        if "default_code_security_configurations_disabled" in included_keys:
            default_configs = await self._get_default_code_security_configurations(org_id)
            settings["default_code_security_configurations_disabled"] = len(default_configs) == 0

        result = {}
        for k, v in settings.items():
            if k in included_keys:
                result[k] = v
                _logger.trace(f"retrieved setting for '{k}' = '{v}'")

        return result

    async def update_settings(self, org_id: str, data: dict[str, Any]) -> None:
        _logger.debug("updating settings for org '%s'", org_id)

        try:
            await self.requester.request_json("PATCH", f"/orgs/{org_id}", data)
        except GitHubException as ex:
            raise RuntimeError(f"failed to update settings for org '{org_id}':\n{ex}") from ex

        if "security_managers" in data:
            await self.update_security_managers(org_id, data["security_managers"])

        if "default_code_security_configurations_disabled" in data:
            default_code_security_configuration_disabled = data["default_code_security_configurations_disabled"]
            if default_code_security_configuration_disabled is True:
                await self._disable_default_code_security_configurations(org_id)
            else:
                _logger.warning("trying to enable default code security configurations")

        _logger.debug("updated %d setting(s)", len(data))

    async def list_security_managers(self, org_id: str, role_id: str) -> list[str]:
        _logger.debug("retrieving security managers for org '%s'", org_id)

        try:
            result = await self.requester.request_json("GET", f"/orgs/{org_id}/organization-roles/{role_id}/teams")
            return [x["slug"] for x in result]
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving security managers for org " f"'{org_id}':\n{ex}") from ex

    async def update_security_managers(self, org_id: str, security_managers: list[str]) -> None:
        _logger.debug("updating security managers for org '%s'", org_id)

        security_manager_role_id = await self.get_role_id(org_id, "security_manager")

        if security_manager_role_id is None:
            _logger.warning(
                "failed to update security managers for org '%s' as role '%s' does not exist yet",
                org_id,
                "security_manager",
            )
            return

        current_managers = set(await self.list_security_managers(org_id, str(security_manager_role_id)))

        # first, add all security managers that are not yet configured.
        for team_slug in security_managers:
            if team_slug in current_managers:
                current_managers.remove(team_slug)
            else:
                await self.assign_role_to_team(org_id, str(security_manager_role_id), team_slug)

        # second, remove the current managers that are left.
        for team_slug in current_managers:
            await self.remove_role_from_team(org_id, str(security_manager_role_id), team_slug)

    async def assign_role_to_team(self, org_id: str, role_id: str, team_slug: str) -> None:
        _logger.debug("assigning role with id '%s' to team '%s' for org '%s'", role_id, team_slug, org_id)

        status, body = await self.requester.request_raw(
            "PUT", f"/orgs/{org_id}/organization-roles/teams/{team_slug}/{role_id}"
        )

        if status == 204:
            _logger.debug("assigned role '%s' to team '%s' for org '%s'", role_id, team_slug, org_id)
        elif status == 404:
            _logger.warning(
                "failed to assign role '%s' to team '%s' for org '%s': %s",
                role_id,
                team_slug,
                org_id,
                body,
            )
        else:
            raise RuntimeError(
                f"failed assigning role '{role_id}' to team '{team_slug}' in org '{org_id}'" f"\n{status}: {body}"
            )

    async def remove_role_from_team(self, org_id: str, role_id: str, team_slug: str) -> None:
        _logger.debug("removing role '%s' from team '%s' in org '%s'", role_id, team_slug, org_id)

        status, body = await self.requester.request_raw(
            "DELETE", f"/orgs/{org_id}/organization-roles/teams/{team_slug}/{role_id}"
        )
        if status != 204:
            raise RuntimeError(
                f"failed removing role '{role_id}' from team '{team_slug}' in org '{org_id}'" f"\n{status}: {body}"
            )

        _logger.debug("removed role '%s' from team '%s' in org '%s'", role_id, team_slug, org_id)

    async def get_role_id(self, org_id: str, role_name: str) -> int | None:
        _logger.debug("retrieving id for role with name '%s' in org '%s'", role_name, org_id)

        try:
            result = await self.requester.request_json("GET", f"/orgs/{org_id}/organization-roles")
            for role in result["roles"]:
                if role["name"] == role_name:
                    return role["id"]

            return None
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving custom roles for org '{org_id}':\n{ex}") from ex

    async def get_custom_roles(self, org_id: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving custom roles for org '%s'", org_id)

        try:
            result = await self.requester.request_json("GET", f"/orgs/{org_id}/organization-roles")
            return [r for r in result["roles"] if r["source"] != "Predefined"]
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving custom roles for org '{org_id}':\n{ex}") from ex

    async def update_custom_role(self, org_id: str, role_id: int, role_name: str, role: dict[str, Any]) -> None:
        _logger.debug("updating org role with name '%s' for org '%s'", role_name, org_id)

        try:
            await self.requester.request_json("PATCH", f"/orgs/{org_id}/organization-roles/{role_id}", role)
            _logger.debug("updated org role with name '%s'", role_name)
        except GitHubException as ex:
            raise RuntimeError(f"failed to update org role '{role_name}':\n{ex}") from ex

    async def add_custom_role(self, org_id: str, role_name: str, data: dict[str, Any]) -> None:
        _logger.debug("adding org role with name '%s' for org '%s'", role_name, org_id)

        try:
            await self.requester.request_json("POST", f"/orgs/{org_id}/organization-roles", data)
            _logger.debug("added org role with name '%s'", role_name)
        except GitHubException as ex:
            raise RuntimeError(f"failed to add org role with name '{role_name}':\n{ex}") from ex

    async def delete_custom_role(self, org_id: str, role_id: int, role_name: str) -> None:
        _logger.debug("deleting org role with name '%s' for org '%s'", role_name, org_id)

        status, _ = await self.requester.request_raw("DELETE", f"/orgs/{org_id}/organization-roles/{role_id}")

        if status != 204:
            raise RuntimeError(f"failed to delete org role with name '{role_name}'")

        _logger.debug("removed org role with name '%s'", role_name)

    async def get_custom_properties(self, org_id: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving custom properties for org '%s'", org_id)

        try:
            return await self.requester.request_json("GET", f"/orgs/{org_id}/properties/schema")
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving custom properties for org '{org_id}':\n{ex}") from ex

    async def add_custom_property(self, org_id: str, property_name: str, data: dict[str, Any]) -> None:
        _logger.debug("adding org custom property with name '%s' for org '%s'", property_name, org_id)

        try:
            await self.requester.request_json("PUT", f"/orgs/{org_id}/properties/schema/{property_name}", data)
            _logger.debug("added org custom property with name '%s'", property_name)
        except GitHubException as ex:
            raise RuntimeError(f"failed to add org custom property with name '{property_name}':\n{ex}") from ex

    async def update_custom_property(self, org_id: str, property_name: str, data: dict[str, Any]) -> None:
        _logger.debug(f"updating org custom property with name '{property_name}'")

        try:
            await self.requester.request_json("PUT", f"/orgs/{org_id}/properties/schema/{property_name}", data)
            _logger.debug("updated org custom property with name '%s'", property_name)
        except GitHubException as ex:
            raise RuntimeError(f"failed to update org custom property with name '{property_name}':\n{ex}") from ex

    async def delete_custom_property(self, org_id: str, property_name: str) -> None:
        _logger.debug("deleting org custom property with name '%s'", property_name)

        status, _ = await self.requester.request_raw("DELETE", f"/orgs/{org_id}/properties/schema/{property_name}")

        if status != 204:
            raise RuntimeError(f"failed to delete org custom property with name '{property_name}'")

        _logger.debug("removed org custom property with name '%s'", property_name)

    async def get_webhook_id(self, org_id: str, url: str) -> str:
        _logger.debug("retrieving id for org webhook with url '%s' for org '%s'", url, org_id)

        webhooks = await self.get_webhooks(org_id)

        has_wildcard_url = url.endswith("*")
        stripped_url = url.rstrip("*")

        for webhook in webhooks:
            webhook_url = webhook["config"]["url"]
            if (has_wildcard_url is True and webhook_url.startswith(stripped_url)) or webhook_url == url:
                return webhook["id"]

        raise RuntimeError(f"failed to find org webhook with url '{url}'")

    async def get_webhooks(self, org_id: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving org webhooks for org '%s'", org_id)

        try:
            return await self.requester.request_json("GET", f"/orgs/{org_id}/hooks")
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving webhooks for org '{org_id}':\n{ex}") from ex

    async def update_webhook(self, org_id: str, webhook_id: int, webhook: dict[str, Any]) -> None:
        _logger.debug("updating org webhook '%d' for org '%s'", webhook_id, org_id)

        try:
            await self.requester.request_json("PATCH", f"/orgs/{org_id}/hooks/{webhook_id}", webhook)
            _logger.debug("updated webhook %d", webhook_id)
        except GitHubException as ex:
            raise RuntimeError(f"failed to update org webhook {webhook_id}:\n{ex}") from ex

    async def add_webhook(self, org_id: str, data: dict[str, Any]) -> None:
        url = data["config"]["url"]
        _logger.debug("adding org webhook with url '%s' for org '%s'", url, org_id)

        # mandatory field "name" = "web"
        data["name"] = "web"

        try:
            await self.requester.request_json("POST", f"/orgs/{org_id}/hooks", data)
            _logger.debug("added org webhook with url '%s'", url)
        except GitHubException as ex:
            raise RuntimeError(f"failed to add org webhook with url '{url}':\n{ex}") from ex

    async def delete_webhook(self, org_id: str, webhook_id: int, url: str) -> None:
        _logger.debug("deleting org webhook with url '%s' for org '%s'", url, org_id)

        status, _ = await self.requester.request_raw("DELETE", f"/orgs/{org_id}/hooks/{webhook_id}")

        if status != 204:
            raise RuntimeError(f"failed to delete org webhook with url '{url}'")

        _logger.debug("removed org webhook with url '%s'", url)

    async def get_repos(self, org_id: str) -> list[str]:
        _logger.debug("retrieving repos for org '%s'", org_id)

        params = {"type": "all"}
        try:
            repos = await self.requester.request_paged_json("GET", f"/orgs/{org_id}/repos", params=params)
            return [repo["name"] for repo in repos]
        except GitHubException as ex:
            raise RuntimeError(f"failed to retrieve repos for org '{org_id}':\n{ex}") from ex

    async def get_secrets(self, org_id: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving secrets for org '%s'", org_id)

        try:
            response = await self.requester.request_json("GET", f"/orgs/{org_id}/actions/secrets")

            secrets = response["secrets"]
            for secret in secrets:
                if secret["visibility"] == "selected":
                    secret["selected_repositories"] = await self._get_selected_repositories_for_secret(
                        org_id, secret["name"]
                    )
            return secrets
        except GitHubException as ex:
            raise RuntimeError(f"failed getting secrets for org '{org_id}':\n{ex}") from ex

    async def _get_selected_repositories_for_secret(self, org_id: str, secret_name: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving selected repositories for secret '%s' in org '%s'", secret_name, org_id)

        try:
            url = f"/orgs/{org_id}/actions/secrets/{secret_name}/repositories"
            response = await self.requester.request_json("GET", url)
            return response["repositories"]
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving selected repositories:\n{ex}") from ex

    async def add_secret(self, org_id: str, data: dict[str, str]) -> None:
        secret_name = data.pop("name")
        _logger.debug("adding org secret '%s' in org '%s'", secret_name, org_id)

        await self._encrypt_secret_inplace(org_id, data)

        status, _ = await self.requester.request_raw(
            "PUT", f"/orgs/{org_id}/actions/secrets/{secret_name}", json.dumps(data)
        )

        if status != 201:
            raise RuntimeError(f"failed to add org secret '{secret_name}'")

        _logger.debug("added org secret '%s'", secret_name)

    async def update_secret(self, org_id: str, secret_name: str, secret: dict[str, Any]) -> None:
        _logger.debug("updating org secret '%s' in org '%s'", secret_name, org_id)

        if "name" in secret:
            secret.pop("name")

        await self._encrypt_secret_inplace(org_id, secret)

        status, _ = await self.requester.request_raw(
            "PUT", f"/orgs/{org_id}/actions/secrets/{secret_name}", json.dumps(secret)
        )

        if status != 204:
            raise RuntimeError(f"failed to update org secret '{secret_name}'")

        _logger.debug("updated org secret '%s'", secret_name)

    async def _encrypt_secret_inplace(self, org_id: str, data: dict[str, Any]) -> None:
        if "value" in data:
            value = data.pop("value")
            key_id, public_key = await self.get_public_key(org_id)
            data["encrypted_value"] = encrypt_value(public_key, value)
            data["key_id"] = key_id

    async def delete_secret(self, org_id: str, secret_name: str) -> None:
        _logger.debug("deleting org secret '%s' in org '%s'", secret_name, org_id)

        status, _ = await self.requester.request_raw("DELETE", f"/orgs/{org_id}/actions/secrets/{secret_name}")
        if status != 204:
            raise RuntimeError(f"failed to delete org secret '{secret_name}'")

        _logger.debug("removed org secret '%s'", secret_name)

    async def get_variables(self, org_id: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving variables for org '%s'", org_id)

        try:
            response = await self.requester.request_json("GET", f"/orgs/{org_id}/actions/variables")

            secrets = response["variables"]
            for secret in secrets:
                if secret["visibility"] == "selected":
                    secret["selected_repositories"] = await self._get_selected_repositories_for_variable(
                        org_id, secret["name"]
                    )
            return secrets
        except GitHubException as ex:
            raise RuntimeError(f"failed getting variables for org '{org_id}':\n{ex}") from ex

    async def _get_selected_repositories_for_variable(self, org_id: str, variable_name: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving selected repositories for variable '%s' in org '%s'", variable_name, org_id)

        try:
            url = f"/orgs/{org_id}/actions/variables/{variable_name}/repositories"
            response = await self.requester.request_json("GET", url)
            return response["repositories"]
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving selected repositories:\n{ex}") from ex

    async def _set_selected_repositories_for_variable(
        self, org_id: str, variable_name: str, selected_repository_ids: list[str]
    ) -> None:
        _logger.debug("setting selected repositories for variable '%s' in org '%s'", variable_name, org_id)

        try:
            data = {"selected_repository_ids": selected_repository_ids}

            url = f"/orgs/{org_id}/actions/variables/{variable_name}/repositories"
            status, _ = await self.requester.request_raw("PUT", url, json.dumps(data))
            if status != 204:
                raise RuntimeError(f"failed to update selected repositories for variable '{variable_name}'")

            _logger.debug("updated selected repositories for variable '%s'", variable_name)

        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving selected repositories:\n{ex}") from ex

    async def update_variable(self, org_id: str, variable_name: str, variable: dict[str, Any]) -> None:
        _logger.debug("updating org variable '%s' in org '%s'", variable_name, org_id)

        if "name" in variable:
            variable.pop("name")

        visibility = variable.get("visibility")

        if "selected_repository_ids" in variable:
            selected_repository_ids = variable.pop("selected_repository_ids")
        else:
            selected_repository_ids = None

        status, body = await self.requester.request_raw(
            "PATCH", f"/orgs/{org_id}/actions/variables/{variable_name}", json.dumps(variable)
        )
        if status != 204:
            raise RuntimeError(f"failed to update org variable '{variable_name}': {body}")

        if selected_repository_ids is not None and (visibility is None or visibility == "selected"):
            await self._set_selected_repositories_for_variable(org_id, variable_name, selected_repository_ids)

        _logger.debug("updated org variable '%s'", variable_name)

    async def add_variable(self, org_id: str, data: dict[str, str]) -> None:
        variable_name = data.get("name")
        _logger.debug("adding org variable '%s' in org '%s'", variable_name, org_id)

        status, body = await self.requester.request_raw("POST", f"/orgs/{org_id}/actions/variables", json.dumps(data))

        if status != 201:
            raise RuntimeError(f"failed to add org variable '{variable_name}': {body}")

        _logger.debug("added org variable '%s'", variable_name)

    async def delete_variable(self, org_id: str, variable_name: str) -> None:
        _logger.debug("deleting org variable '%s' in org '%s'", variable_name, org_id)

        status, body = await self.requester.request_raw("DELETE", f"/orgs/{org_id}/actions/variables/{variable_name}")

        if status != 204:
            raise RuntimeError(f"failed to delete org variable '{variable_name}': {body}")

        _logger.debug("removed org variable '%s'", variable_name)

    async def get_public_key(self, org_id: str) -> tuple[str, str]:
        _logger.debug("retrieving org public key for org '%s'", org_id)

        try:
            response = await self.requester.request_json("GET", f"/orgs/{org_id}/actions/secrets/public-key")
            return response["key_id"], response["key"]
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving org public key:\n{ex}") from ex

    async def get_app_installations(self, org_id: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving app installations for org '%s'", org_id)

        try:
            response = await self.requester.request_json("GET", f"/orgs/{org_id}/installations")
            return response["installations"]
        except GitHubException as ex:
            raise RuntimeError(f"failed getting app installations for org '{org_id}':\n{ex}") from ex

    async def get_workflow_settings(self, org_id: str) -> dict[str, Any]:
        _logger.debug("retrieving workflow settings for org '%s'", org_id)

        workflow_settings: dict[str, Any] = {}

        try:
            permissions = await self.requester.request_json("GET", f"/orgs/{org_id}/actions/permissions")
            workflow_settings.update(permissions)
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving workflow settings for org '{org_id}':\n{ex}") from ex

        if permissions["enabled_repositories"] == "selected":
            workflow_settings["selected_repositories"] = await self._get_selected_repositories_for_workflow_settings(
                org_id
            )
        else:
            workflow_settings["selected_repositories"] = None

        allowed_actions = permissions.get("allowed_actions", "none")
        if allowed_actions == "selected":
            workflow_settings.update(await self._get_selected_actions_for_workflow_settings(org_id))

        if allowed_actions != "none":
            workflow_settings.update(await self._get_default_workflow_permissions(org_id))

        return workflow_settings

    async def update_workflow_settings(self, org_id: str, data: dict[str, Any]) -> None:
        _logger.debug("updating workflow settings for org '%s'", org_id)

        permission_data = {k: data[k] for k in ["enabled_repositories", "allowed_actions"] if k in data}
        if len(permission_data) > 0:
            status, body = await self.requester.request_raw(
                "PUT", f"/orgs/{org_id}/actions/permissions", json.dumps(permission_data)
            )

            if status != 204:
                raise RuntimeError(f"failed to update workflow settings for org '{org_id}'" f"\n{status}: {body}")

            _logger.debug("updated workflow settings for org '%s'", org_id)

        if "selected_repository_ids" in data:
            await self._update_selected_repositories_for_workflow_settings(org_id, data["selected_repository_ids"])

        allowed_action_data = {
            k: data[k] for k in ["github_owned_allowed", "verified_allowed", "patterns_allowed"] if k in data
        }
        if len(allowed_action_data) > 0:
            await self._update_selected_actions_for_workflow_settings(org_id, allowed_action_data)

        default_permission_data = {
            k: data[k] for k in ["default_workflow_permissions", "can_approve_pull_request_reviews"] if k in data
        }
        if len(default_permission_data) > 0:
            await self._update_default_workflow_permissions(org_id, default_permission_data)

        _logger.debug("updated %d workflow setting(s)", len(data))

    async def _get_selected_repositories_for_workflow_settings(self, org_id: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving selected repositories for org workflow settings of org '%s'", org_id)

        try:
            response = await self.requester.request_json("GET", f"/orgs/{org_id}/actions/permissions/repositories")
            return response["repositories"]
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving selected repositories:\n{ex}") from ex

    async def _update_selected_repositories_for_workflow_settings(
        self, org_id: str, selected_repository_ids: list[int]
    ) -> None:
        _logger.debug("updating selected repositories for org workflow settings of org '%s'", org_id)

        data = {"selected_repository_ids": selected_repository_ids}
        status, body = await self.requester.request_raw(
            "PUT", f"/orgs/{org_id}/actions/permissions/repositories", json.dumps(data)
        )

        if status != 204:
            raise RuntimeError(
                f"failed updating selected repositories for workflow settings of org '{org_id}'" f"\n{status}: {body}"
            )

        _logger.debug("updated selected repositories for workflow settings of org '%s'", org_id)

    async def _get_selected_actions_for_workflow_settings(self, org_id: str) -> dict[str, Any]:
        _logger.debug("retrieving allowed actions for org '%s'", org_id)

        try:
            return await self.requester.request_json("GET", f"/orgs/{org_id}/actions/permissions/selected-actions")
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving allowed actions for org '{org_id}':\n{ex}") from ex

    async def _update_selected_actions_for_workflow_settings(self, org_id: str, data: dict[str, Any]) -> None:
        _logger.debug("updating allowed actions for org '%s'", org_id)

        status, body = await self.requester.request_raw(
            "PUT", f"/orgs/{org_id}/actions/permissions/selected-actions", json.dumps(data)
        )

        if status != 204:
            raise RuntimeError(f"failed updating allowed actions for org '{org_id}'" f"\n{status}: {body}")

        _logger.debug("updated allowed actions for org '%s'", org_id)

    async def _get_default_workflow_permissions(self, org_id: str) -> dict[str, Any]:
        _logger.debug("retrieving default workflow permissions for org '%s'", org_id)

        try:
            return await self.requester.request_json("GET", f"/orgs/{org_id}/actions/permissions/workflow")
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving org default workflow permissions:\n{ex}") from ex

    async def _update_default_workflow_permissions(self, org_id: str, data: dict[str, Any]) -> None:
        _logger.debug("updating default workflow permissions for org '%s'", org_id)

        status, body = await self.requester.request_raw(
            "PUT", f"/orgs/{org_id}/actions/permissions/workflow", json.dumps(data)
        )

        if status != 204:
            raise RuntimeError(f"failed updating default workflow permissions for org '{org_id}'" f"\n{status}: {body}")

        _logger.debug("updated default workflow permissions for org '%s'", org_id)

    async def _get_default_code_security_configurations(self, org_id: str) -> list[dict[str, Any]]:
        _logger.debug(f"retrieving default code security configurations for org '{org_id}'")

        try:
            url = f"/orgs/{org_id}/code-security/configurations/defaults"
            return await self.requester.request_json("GET", url)
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving default code security configurations:\n{ex}") from ex

    async def _disable_default_code_security_configurations(self, org_id: str) -> None:
        _logger.debug("retrieving default code security configurations for org '%s'", org_id)

        current_default_configs = await self._get_default_code_security_configurations(org_id)

        for config in current_default_configs:
            configuration_id = config["configuration"]["id"]

            try:
                data = {"default_for_new_repos": "none"}
                url = f"/orgs/{org_id}/code-security/configurations/{configuration_id}/defaults"
                await self.requester.request_json("PUT", url, data=data)
            except GitHubException as ex:
                raise RuntimeError(
                    f"failed disabling default code security configuration with id {configuration_id}:\n{ex}"
                ) from ex

    async def list_members(self, org_id: str, two_factor_disabled: bool = False) -> list[dict[str, Any]]:
        _logger.debug("retrieving list of org members for org '%s'", org_id)

        try:
            params = {"filter": "2fa_disabled"} if two_factor_disabled is True else None
            return await self.requester.request_paged_json("GET", f"/orgs/{org_id}/members", params=params)
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving members:\n{ex}") from ex

    async def get_security_advisories(self, org_id: str, state: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving security advisories for org '%s' with state '%s'", org_id, state)

        try:
            params = {"state": state} if state is not None else None
            return await self.requester.request_paged_json("GET", f"/orgs/{org_id}/security-advisories", params=params)
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving security advisories for org '{org_id}':\n{ex}") from ex

    async def get_rulesets(self, org_id: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving org rulesets for org '%s'", org_id)

        try:
            result = []
            response = await self.requester.request_paged_json("GET", f"/orgs/{org_id}/rulesets")
            for ruleset in response:
                result.append(await self.get_ruleset(org_id, str(ruleset["id"])))
            return result
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving org rulesets for org '{org_id}':\n{ex}") from ex

    async def get_ruleset(self, org_id: str, ruleset_id: str) -> dict[str, Any]:
        _logger.debug("retrieving org ruleset '%s' for org '%s'", ruleset_id, org_id)

        try:
            return await self.requester.request_json("GET", f"/orgs/{org_id}/rulesets/{ruleset_id}")
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving org ruleset for org '{org_id}':\n{ex}") from ex

    async def get_ruleset_id(self, org_id: str, name: str) -> str:
        _logger.debug("retrieving id for org ruleset with name '%s' for org '%s'", name, org_id)

        rulesets = await self.get_rulesets(org_id)

        for ruleset in rulesets:
            if ruleset["name"] == name:
                return ruleset["id"]

        raise RuntimeError(f"failed to find org ruleset with name '{name}'")

    async def update_ruleset(self, org_id: str, ruleset_id: int, ruleset: dict[str, Any]) -> None:
        _logger.debug("updating org ruleset '%d' for org '%s'", ruleset_id, org_id)

        try:
            await self.requester.request_json("PUT", f"/orgs/{org_id}/rulesets/{ruleset_id}", ruleset)
            _logger.debug("updated org ruleset '%d'", ruleset_id)
        except GitHubException as ex:
            raise RuntimeError(f"failed to update org ruleset {ruleset_id}:\n{ex}") from ex

    async def add_ruleset(self, org_id: str, data: dict[str, Any]) -> None:
        name = data["name"]
        _logger.debug("adding org ruleset with name '%s' for org '%s'", name, org_id)

        try:
            await self.requester.request_json("POST", f"/orgs/{org_id}/rulesets", data)
            _logger.debug(f"added org ruleset with name '{name}'")
        except GitHubException as ex:
            raise RuntimeError(f"failed to add org ruleset with name '{name}':\n{ex}") from ex

    async def delete_ruleset(self, org_id: str, ruleset_id: int, name: str) -> None:
        _logger.debug("deleting org ruleset with name '%s' for org '%s'", name, org_id)

        status, _ = await self.requester.request_raw("DELETE", f"/orgs/{org_id}/rulesets/{ruleset_id}")

        if status != 204:
            raise RuntimeError(f"failed to delete org ruleset with name '{name}'")

        _logger.debug("removed org ruleset with name '%s'", name)



#  *******************************************************************************
#  Copyright (c) 2023-2025 Eclipse Foundation and others.
#  This program and the accompanying materials are made available
#  under the terms of the Eclipse Public License 2.0
#  which is available at http://www.eclipse.org/legal/epl-v20.html
#  SPDX-License-Identifier: EPL-2.0
#  *******************************************************************************

from typing import Any

from otterdog.logging import get_logger
from otterdog.providers.github.exception import GitHubException

from . import RestApi, RestClient

_logger = get_logger(__name__)


class ActionClient(RestClient):
    def __init__(self, rest_api: RestApi):
        super().__init__(rest_api)

    async def get_workflows(self, org_id: str, repo: str) -> list[dict[str, Any]]:
        _logger.debug("retrieving workflows for repo '%s/%s'", org_id, repo)

        try:
            return await self.requester.request_paged_json(
                "GET", f"/repos/{org_id}/{repo}/actions/workflows", entries_key="workflows"
            )
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving workflows for '{org_id}/{repo}':\n{ex}") from ex

    async def cancel_workflow_run(self, org_id: str, repo_name: str, run_id: str) -> bool:
        _logger.debug("cancelling workflow run #%s in repo '%s/%s'", run_id, org_id, repo_name)

        status, body = await self.requester.request_raw(
            "POST", f"/repos/{org_id}/{repo_name}/actions/runs/{run_id}/cancel"
        )

        if status == 202:
            return True
        elif status == 409:
            return False
        else:
            raise RuntimeError(
                f"failed cancelling workflow run #{run_id} in repo '{org_id}/{repo_name}'\n{status}: {body}"
            )

    async def get_artifacts(self, org_id: str, repo: str, run_id: int) -> list[dict[str, Any]]:
        _logger.debug("list artifacts for workflow run #%d in repo '%s/%s'", run_id, org_id, repo)

        try:
            return await self.requester.request_paged_json(
                "GET", f"/repos/{org_id}/{repo}/actions/runs/{run_id}/artifacts", entries_key="artifacts"
            )
        except GitHubException as ex:
            raise RuntimeError(f"failed retrieving artifacts for '{org_id}/{repo}':\n{ex}") from ex

    async def download_artifact(self, file, org_id: str, repo_name: str, artifact_id: int) -> None:
        _logger.debug("downloading workflow artifact in repo '%s/%s'", org_id, repo_name)

        try:
            async for data in self.requester.request_stream(
                "GET", f"/repos/{org_id}/{repo_name}/actions/artifacts/{artifact_id}/zip"
            ):
                await file.write(data)

        except GitHubException as ex:
            raise RuntimeError(f"failed downloading workflow artifact from repo '{org_id}/{repo_name}':\n{ex}") from ex
